# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Folium map class."""

import contextlib
import itertools
import math
import statistics as stats
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import folium
import pandas as pd
from folium.plugins import FeatureGroupSubGroup, MarkerCluster

from .._version import VERSION
from ..common.exceptions import MsticpyMissingDependencyError
from ..common.utility import export
from ..datamodel.entities import Entity, GeoLocation, IpAddress

from ..context.geoip import GeoLiteLookup  # isort: skip

try:
    import pygeohash
except ImportError:
    pygeohash = None  # pylint: disable=invalid-name


__version__ = VERSION
__author__ = "Ian Hellen"

_GEO_LITE = GeoLiteLookup()


# pylint: disable=too-many-lines
@export
class FoliumMap:
    """Wrapper class for Folium/Leaflet mapping."""

    def __init__(
        self,
        title: str = "OpenStreetMap",
        zoom_start: float = 2.5,
        tiles=None,
        width: str = "100%",
        height: str = "100%",
        location: list = None,
    ):
        """
        Create an instance of the folium map.

        Parameters
        ----------
        title : str, optional
            Name of the main tile layer (the default is 'OpenStreetMap')
        zoom_start : int, optional
            The zoom level of the map (the default is 7)
        tiles : [type], optional
            Custom set of tiles or tile URL (the default is None)
        width : str, optional
            Map display width (the default is '100%')
        height : str, optional
            Map display height (the default is '100%')
        location : list, optional
            Location to center map on

        Attributes
        ----------
        folium_map : folium.Map
            The map object.

        """
        if not location:
            location = [47.67, -122.13]

        self.folium_map = folium.Map(
            zoom_start=zoom_start,
            tiles=tiles,
            width=width,
            height=height,
            location=location,
            control_scale=True,
        )
        folium.TileLayer(name=title).add_to(self.folium_map)
        self.locations: List[Tuple[float, float]] = []

    def _repr_html_(self):
        """Return folium map as HTML."""
        # pylint: disable=protected-access
        return self.folium_map._repr_html_()
        # pylint: enable=protected-access

    def center_map(self):
        """Calculate and set map center based on current coordinates."""
        self.folium_map.location = _get_center_coords(self.locations)

    def add_ip_cluster(
        self, ip_entities: Iterable[IpAddress], layer: Optional[str] = None, **kwargs
    ):
        """
        Add a collection of IP Entities to the map.

        Parameters
        ----------
        ip_entities : Iterable[IpAddress]
            a iterable of IpAddress Entities
        layer : str, optional
            If not none, it will add the entities to a new layer.

        Other Parameters
        ----------------
        kwargs: icon properties to use for displaying this cluster

        """
        ip_entities = _get_location_for_ip_entities(ip_entities)

        if layer:
            marker_target = folium.FeatureGroup(name=layer)
            marker_target.add_to(self.folium_map)
            folium.LayerControl().add_to(self.folium_map)
        else:
            marker_target = self.folium_map
        for ip_entity in ip_entities:
            if ip_entity.Location is None:
                continue
            if (
                not (
                    isinstance(ip_entity.Location.Latitude, (int, float))
                    and isinstance(ip_entity.Location.Longitude, (int, float))
                )
                or math.isnan(ip_entity.Location.Latitude)
                or math.isnan(ip_entity.Location.Longitude)
            ):
                continue
            popup_text = _get_popup_text(ip_entity)
            tooltip_text = _get_tooltip_text(ip_entity)

            marker = folium.Marker(
                location=[ip_entity.Location.Latitude, ip_entity.Location.Longitude],
                popup=popup_text,
                tooltip=tooltip_text,
                icon=folium.Icon(**kwargs),
            )
            marker.add_to(marker_target)
            self.locations.append(
                (ip_entity.Location.Latitude, ip_entity.Location.Longitude)
            )

    def add_ips(self, ip_addresses: Iterable[str], **kwargs):
        """
        Add a collection of GeoLocation objects to the map.

        Parameters
        ----------
        ip_addresses : Iterable[str]
            Iterable of ip strings.
        layer : str, optional
            If not none, it will add the entities to a new layer.

        Other Parameters
        ----------------
        kwargs: icon properties to use for displaying this cluster

        """
        _, ip_entities = _GEO_LITE.lookup_ip(ip_addr_list=ip_addresses)
        self.add_ip_cluster(ip_entities=ip_entities, **kwargs)

    def add_geoloc_cluster(self, geo_locations: Iterable[GeoLocation], **kwargs):
        """
        Add a collection of GeoLocation objects to the map.

        Parameters
        ----------
        geo_locations : Iterable[GeoLocation]
            Iterable of GeoLocation entities.
        layer : str, optional
            If not none, it will add the entities to a new layer.

        Other Parameters
        ----------------
        kwargs: icon properties to use for displaying this cluster

        """
        ip_entities = [IpAddress(Address="na", Location=geo) for geo in geo_locations]
        self.add_ip_cluster(ip_entities=ip_entities, **kwargs)

    def add_locations(self, locations: Iterable[Tuple[float, float]], **kwargs):
        """
        Add a collection of lat/long tuples to the map.

        Parameters
        ----------
        locations : Iterable[Tuple[float, float]]
            Iterable of location tuples.
        layer : str, optional
            If not none, it will add the entities to a new layer.

        Other Parameters
        ----------------
        kwargs: icon properties to use for displaying this cluster

        """
        geo_entities = [
            GeoLocation(Latitude=lat, Longitude=long) for lat, long in locations
        ]
        self.add_geoloc_cluster(geo_locations=geo_entities, **kwargs)

    def add_geo_hashes(self, geohashes: Iterable[str], **kwargs):
        """
        Add decoded geohashes to the map.

        Parameters
        ----------
        geohashes : Iterable[str]
            Iterable of geolocation hashes
        layer : str, optional
            If not none, it will add the entities to a new layer.

        Other Parameters
        ----------------
        kwargs: icon properties to use for displaying this cluster

        """
        geo_entities = []
        for geohash in geohashes:
            decoded_location = decode_geo_hash(geohash)
            geo_entities.append(
                GeoLocation(Latitude=decoded_location[0], Longitude=decoded_location[1])
            )

        self.add_geoloc_cluster(geo_locations=geo_entities, **kwargs)

    def add_marker_clusters(self, clusters: Iterable[MarkerCluster]):
        """
        Add MarkerClusters and to the map.

        Parameters
        ----------
        clusters: Iterable[MarkerCluster]
            Iterable of MarkerClusters

        """
        for cluster in clusters:
            self.folium_map.add_child(cluster)

    def add_feature_sub_groups(self, subgroups: Iterable[FeatureGroupSubGroup]):
        """
        Add FeatureGroupSubGroups and to the map.

        Parameters
        ----------
        subgroups: Iterable[FeatureGroupSubGroup]
            Iterable of FeatureGroupSubGroups

        """
        for subgroup in subgroups:
            self.folium_map.add_child(subgroup)

    def save_map(self, path: str):
        """
        Save the map to `path`.

        Parameters
        ----------
        path: str
            File path to save the current map

        """
        self.folium_map.save(path)

    def add_locations_to_feature_subgroup(
        self,
        locations: Iterable[Tuple[float, float]],
        subgroup: FeatureGroupSubGroup,
        **kwargs,
    ):
        """
        Create markers from locations and add the FeatureGroupSubGroup.

        Parameters
        ----------
        locations: Iterable[Tuple[float, float]]
            Collection of Latitude/Longitude coordinates to be added
            to the FeatureGroupSubGroup
        subgroup: FeatureGroupSubGroup
            Subgroup to add locations to, then add to the map

        """
        for point in locations:
            marker = self.create_marker(location=point, **kwargs)
            marker.add_to(subgroup)
            self.locations.append(point)

        subgroups = [subgroup]

        self.add_feature_sub_groups(subgroups)

    def add_locations_to_marker_cluster(
        self, locations: Iterable[Tuple[float, float]], cluster: MarkerCluster, **kwargs
    ):
        """
        Create markers from locations and add to MarkerCluster.

        Parameters
        ----------
        locations: Iterable[Tuple[float, float]]
            Collection of Latitude/Longitude coordinates to be added
            to the MarkerCluster
        cluster: MarkerCluster
            Marker cluster to add locations to, then add to the map

        """
        for point in locations:
            marker = self.create_marker(location=point, **kwargs)
            marker.add_to(cluster)
            self.locations.append(point)

        clusters = [cluster]

        self.add_marker_clusters(clusters)

    def create_new_cluster_with_locations(
        self, locations: Iterable[Tuple[float, float]], name: str, **kwargs
    ):
        """
        Create a MarkerCluster with locations.

        Parameters
        ----------
        locations: Iterable[Tuple[float, float]]
            Collection of Latitude/Longitude coordinates to be added to the
            MarkerCluster
        name: str
            Name of Marker Cluster to create, add locations to,
            then add to the map

        """
        marker_cluster = MarkerCluster(name=name)

        self.add_locations_to_marker_cluster(
            locations=locations, cluster=marker_cluster, **kwargs
        )

    def create_new_subgroup_with_locations(
        self,
        locations: Iterable[Tuple[float, float]],
        subgroup_name: str,
        cluster_name: str,
        **kwargs,
    ):
        """
        Create subgroup of markers from locations.

        Parameters
        ----------
        locations: Iterable[Tuple[float, float]]
            Collection of Latitude/Longitude coordinates to be added
            to the FeatureGroupSubGroup
        subgroup_name: str
            Name of FeatureGroupSubGroup to create, add locations to,
            then add to the map
        cluster_name : str
            Name of the cluster

        Notes
        -----
        This function creates a marker cluster and FeatureGroupSubGroup,
        then add the locations to the subgroup, then add the subgroup to the map.

        """
        marker_cluster = MarkerCluster(name=cluster_name)
        feature_subgroup = FeatureGroupSubGroup(marker_cluster, name=subgroup_name)

        self.add_locations_to_feature_subgroup(
            locations=locations, subgroup=feature_subgroup, **kwargs
        )

    def enable_layer_control(self):
        """
        Enable Layer Control on the map.

        Parameters
        ----------
        None

        """
        folium.LayerControl().add_to(self.folium_map)

    def create_new_cluster_with_geohashes(
        self, geohashes: Iterable[str], name: str, **kwargs
    ):
        """
        Create a MarkerCluster and add geohash locations.

        Parameters
        ----------
        geohashes: Iterable[str]
            Collection of geohashes to be decoded and added to the MarkerCluster
        name: str
            Name of Marker Cluster to create, add locations to, then add to the map

        """
        locations = decode_geohash_collection(geohashes)
        self.create_new_cluster_with_locations(locations=locations, name=name, **kwargs)

    def create_new_subgroup_with_geohashes(
        self, geohashes: Iterable[str], subgroup_name: str, cluster_name: str, **kwargs
    ):
        """
        Create a FeatureSubGroup with collection of geohash locations.

        Parameters
        ----------
        geohashes: Iterable[str]
            Collection of geohashes to be decoded and added to
            the FeatureGroupSubGroup
        subgroup_name: str
            Name of SubGroup to create, add locations to, then add to the map
        cluster_name: str
            Name of the Marker Cluster to create and add the SubGroup to

        """
        locations = decode_geohash_collection(geohashes)

        self.create_new_subgroup_with_locations(
            locations=locations,
            subgroup_name=subgroup_name,
            cluster_name=cluster_name,
            **kwargs,
        )

    @staticmethod
    def create_marker(
        location: Tuple[float, float],
        tooltip: str = None,
        popup: str = None,
        **kwargs,
    ) -> folium.Marker:
        """
        Create and return a Folium Marker at a given location.

        Parameters
        ----------
        location: Tuple[float,float]
            Latitude/Longitude coordinates for the Marker
        tooltip: str [Optional]
            Tooltip text for the Marker
        popup: str [Optional]
            Popup text for the Marker

        Returns
        -------
        Marker
            A Folium Marker at the given location coordinates

        """
        return folium.Marker(
            location=location, tooltip=tooltip, popup=popup, icon=folium.Icon(**kwargs)
        )

    @staticmethod
    def create_marker_cluster(name: str):
        """
        Create and return a MarkerCluster with name.

        Parameters
        ----------
        name: str
            Name of the MarkerCluster

        Returns
        -------
        MarkerCluster
            A Folium MarkerCluster with the provided name

        """
        return MarkerCluster(name=name)

    @staticmethod
    def create_feature_sub_group_of_marker_cluster(
        cluster: MarkerCluster, name: str
    ) -> FeatureGroupSubGroup:
        """
        Return a FeatureGroupSubGroup with `name` for a MarkerCluster.

        Parameters
        ----------
        cluster: MarkerCluster
            Folium MarkerCluster to add FeatureGroupSubGroup to
        name: str
            Desired name of the MarkerCluster

        Returns
        -------
        FeatureGroupSubGroup
            A Folium FeatureGroupSubGroup with the provided name as part
            of the given MarkerCluster

        """
        return FeatureGroupSubGroup(cluster, name=name)


@export
def decode_geo_hash(geohash: str) -> Tuple[float, float, float, float]:
    """
    Decode a geohash.

    Parameters
    ----------
    geohash: str
        A string representation of a location

    Returns
    -------
    Tuple
        Tuple representation of a geohash, format of:
        (Latitude, Longitude,
        Latitude Error interval, Longitude Error Interval)

    Raises
    ------
    MsticpyMissingDependencyError
        If pygeohash is not installed.

    """
    if pygeohash is not None:
        return pygeohash.decode_exactly(geohash)
    raise MsticpyMissingDependencyError(packages="pygeohash")


@export
def decode_geohash_collection(geohashes: Iterable[str]):
    """
    Return collection of geohashes decoded into location coordinates.

    Parameters
    ----------
    geohashes: Iterable[str]
        Collection of geohashes to be decoded

    Returns
    -------
    Iterable[Tuple[float, float]]
        Collection of location coordinates in Latitude/Longitude

    """
    locations = []

    for geohash in geohashes:
        exact_location = decode_geo_hash(geohash)
        locations.append((exact_location[0], exact_location[1]))

    return locations


def _get_tooltip_text(ip_entity):
    """Return tooltip text for marker."""
    return "<br>".join(
        str(line)
        for line in [
            ip_entity.Address,
            ip_entity.Location.City or "Unknown city",
            ip_entity.Location.CountryCode or "Unknown country",
            *(list(ip_entity.AdditionalData.items())),
        ]
    )


def _get_popup_text(ip_entity):
    """Return popup text for marker."""
    return "<br>".join(
        str(line)
        for line in [
            ip_entity.Address,
            *(list(ip_entity.Location.properties.values())),
            *(list(ip_entity.AdditionalData.items())),
        ]
    )


IconMapper = Union[Callable[[str], Dict[str, Any]], Dict[str, Any], None]


# pylint: disable=too-many-locals, too-many-arguments
def plot_map(
    data: pd.DataFrame,
    ip_column: Optional[str] = None,
    lat_column: Optional[str] = None,
    long_column: Optional[str] = None,
    layer_column: Optional[str] = None,
    icon_column: Optional[str] = None,
    icon_map: IconMapper = None,
    popup_columns: Optional[List[str]] = None,
    tooltip_columns: Optional[List[str]] = None,
    **kwargs,
) -> folium.Map:
    """
    Plot folium map from DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame, must have either an IP address
        column or latitude and longitude columns.
    ip_column : Optional[str], optional
        The name of the IP Address column, by default None
    lat_column : Optional[str], optional
        The name of the location 'latitude' column, by default None
    long_column : Optional[str], optional
        The name of the location 'longitude' column, by default None
    layer_column : Optional[str], optional
        The column to group markers into for displaying on different
        map layers, by default None
    icon_column : Optional[str], optional
        Optional column containing the name of the icon to use
        for the marker in this row, by default None
    icon_map : IconMapper, optional
        Mapping dictionary or function, by default None
        See Notes for more details.
    popup_columns : Optional[List[str]], optional
        List of columns to use for the popup text, by default None
    tooltip_columns : Optional[List[str]], optional
        List of columns to use for the tooltip text, by default None


    Other Parameters
    ----------------
    marker_cluster : bool, optional
        Use marker clustering, default is True.
    default_color : str, optional
        Default color for marker icons, by default "blue"
    title : str, optional
        Name of the layer (the default is 'layer1')
        (passed to FoliumMap constructor)
    zoom_start : int, optional
        The zoom level of the map (the default is 7)
        (passed to FoliumMap constructor)
    tiles : [type], optional
        Custom set of tiles or tile URL (the default is None)
        (passed to FoliumMap constructor)
    width : str, optional
        Map display width (the default is '100%')
        (passed to FoliumMap constructor)
    height : str, optional
        Map display height (the default is '100%')
        (passed to FoliumMap constructor)
    location : list, optional
        Location to center map on

    Returns
    -------
    folium.Map
        Folium Map object.

    Raises
    ------
    ValueError
        If neither `ip_column` nor `lat_column` and `long_column` are passed.
    LookupError
        If one of the passed columns does not exist in `data`

    Notes
    -----
    There are two ways of providing custom icon settings based on the
    the row of the input DataFrame.

    If `icon_map` is a dict it should contain keys that map to the
    value of `icon_col` and values that a dicts of valid
    folium Icon properties ("color", "icon_color", "icon", "angle", "prefix").
    The dict should include a "default" entry that will be used if the
    value in the DataFrame[icon_col] doesn't match any key.
    For example:

    .. code:: python

        icon_map = {
            "high": {
                "color": "red",
                "icon": "warning",
            },
            "medium": {
                "color": "orange",
                "icon": "triangle-exclamation",
                "prefix": "fa",
            },
            "default": {
                "color": "blue",
                "icon": "info-sign",
            },
        }

    If icon_map is a function it should take a single str parameter
    (the item key) and return a dict of icon properties. It should
    return a default set of values if the key does not match a known
    key. The `icon_col` value for each row will be passed to this
    function and the return value used to populate the Icon arguments.

    For example:

    .. code::python

        def icon_mapper(icon_key):
            if icon_key.startswith("bad"):
                return {
                    "color": "red",
                    "icon": "triangle-alert",
                }
            ...
            else:
                return {
                    "color": "blue",
                    "icon": "info-sign",
                }

    FontAwesome icon (prefix "fa") names are available at https://fontawesome.com/
    GlyphIcons icons (prefix "glyphicon") are available at https://www.glyphicons.com/

    """
    folium_map = FoliumMap(**kwargs)
    if ip_column and not (lat_column and long_column):
        # resolve IP location and merge with input data.
        data = data.merge(
            # pylint: disable=no-member
            _GEO_LITE.lookup_ips(data, column=ip_column),
            # pylint: enable=no-member
            left_on=ip_column,
            right_on="IpAddress",
            suffixes=("_src", None),
        ).dropna(axis="index", subset=["Latitude", "Longitude"])
        lat_column, long_column = ["Latitude", "Longitude"]
        if not tooltip_columns:
            tooltip_columns = _default_columns(data, [ip_column, "CountryCode", "City"])
        if not popup_columns:
            popup_columns = _default_columns(
                data,
                [ip_column, "CountryOrRegionName", "City", lat_column, long_column],
            )
    else:
        if not tooltip_columns:
            tooltip_columns = []
            tooltip_columns.extend(col for col in (lat_column, long_column) if col)
        if not popup_columns:
            popup_columns = []
            popup_columns.extend(col for col in (lat_column, long_column) if col)

    _validate_columns(
        data,
        ip_column,
        lat_column,
        long_column,
        [layer_column, icon_column],
    )
    popup_columns = _validate_optional_columns(data, popup_columns)
    tooltip_columns = _validate_optional_columns(data, tooltip_columns)

    folium_map.locations.extend(
        data.apply(lambda row: (row[lat_column], row[long_column]), axis=1)
    )
    # common dictionary of kwargs to _create_feature_group
    static_kwargs = {
        "lat_column": lat_column,
        "long_column": long_column,
        "icon_column": icon_column,
        "icon_map": icon_map,
        "popup_cols": popup_columns,
        "tooltip_cols": tooltip_columns,
        "use_marker_cluster": kwargs.pop("marker_cluster", True),
    }
    if layer_column is None:
        feature_group = _create_feature_group(
            data=data,
            layer_name="All locations",
            def_layer_color=kwargs.pop("default_color", "blue"),
            **static_kwargs,  # type: ignore
        )
        feature_group.add_to(folium_map.folium_map)
    else:
        for index, (layer, layer_df) in enumerate(data.groupby(layer_column)):
            def_layer_color = _get_icon_layer_color(index)
            feature_group = _create_feature_group(
                data=layer_df,
                layer_name=layer,  # type: ignore
                def_layer_color=def_layer_color,
                **static_kwargs,  # type: ignore
            )
            feature_group.add_to(folium_map.folium_map)
        folium.LayerControl().add_to(folium_map.folium_map)
    if "location" not in kwargs:
        folium_map.center_map()
    return folium_map


# pylint: enable=too-many-locals, too-many-arguments


def _default_columns(data, defaults: Iterable[str]) -> List[str]:
    return [col_name for col_name in defaults if col_name in data.columns]


def _validate_columns(data, ip_column, lat_column, long_column, other_columns):
    """Validate required columns and that optional cols are in the data."""
    if not ip_column and not (lat_column and long_column):
        raise ValueError(
            "Data must have either an IpAddress ('ip_column')",
            "or latitude ('lat_column') and longitude ('long_column')",
        )
    param_cols: List[str] = []
    for param in other_columns:
        if not param:
            continue
        if isinstance(param, list):
            param_cols.extend(param)
        else:
            param_cols.append(param)
    missing_columns = {col for col in param_cols if col not in data.columns}
    if missing_columns:
        raise LookupError(
            "The following columns are not in the supplied DataFrame",
            ",".join(f"'{col}'" for col in missing_columns),
        )


def _validate_optional_columns(data, optional_columns: Iterable[str]) -> List[str]:
    """Validate that optional columns are in the data."""
    return [col for col in optional_columns if col in data.columns]


# pylint: disable=too-many-arguments
def _create_feature_group(
    data: pd.DataFrame,
    layer_name: str,
    lat_column: str,
    long_column: str,
    icon_column: Optional[str],
    icon_map: IconMapper,
    popup_cols: List[str],
    tooltip_cols: List[str],
    def_layer_color: str,
    use_marker_cluster: bool = True,
) -> folium.FeatureGroup:
    """Create folium feature group."""
    feature_group = folium.FeatureGroup(name=layer_name)
    if use_marker_cluster:
        container = MarkerCluster(name=layer_name)
        container.add_to(feature_group)
    else:
        container = feature_group
    data.apply(
        lambda row: folium.Marker(
            location=(row[lat_column], row[long_column]),
            tooltip=_create_marker_text(row, tooltip_cols),
            popup=_create_marker_text(row, popup_cols),
            icon=_create_mapped_icon(row, icon_column, icon_map, def_layer_color),
        ).add_to(feature_group),
        axis=1,
    )
    return feature_group


# pylint: enable=too-many-arguments


def _create_marker_text(row: pd.Series, columns: List[str]) -> str:
    """Return HTML formatted text for tooltips and popups."""
    return "<br>".join(f"{col}: {row[col]}" for col in columns)


def _get_icon_layer_color(layer_index: int) -> str:
    """Get a color from folium.color options."""
    col_options = folium.Icon.color_options - {"white"}
    return list(col_options)[layer_index % len(col_options)]


def _create_mapped_icon(
    row: pd.Series,
    icon_column: Optional[str] = None,
    icon_map: Optional[IconMapper] = None,
    def_layer_color: str = "blue",
) -> folium.Icon:
    """Return folium Icon from mapping or defaults."""
    icon_kwargs: Dict[str, str] = {}
    if isinstance(icon_map, dict):
        icon_kwargs = icon_map.get(row[icon_column], icon_map.get("default", {}))  # type: ignore
    elif callable(icon_map):
        icon_kwargs = icon_map(row[icon_column])  # type: ignore
    elif icon_column:
        icon_kwargs = {"icon": row[icon_column]}
    if "color" not in icon_kwargs:
        icon_kwargs["color"] = def_layer_color
    return folium.Icon(**icon_kwargs)


def _get_location_for_ip_entities(
    ip_entities: Iterable[IpAddress],
) -> Generator[IpAddress, None, None]:
    for ip_entity in ip_entities:
        if (
            ip_entity.Location is None
            or ip_entity.Location.Longitude is None
            or ip_entity.Location.Latitude is None
        ):
            _, ip_res_list = _GEO_LITE.lookup_ip(ip_entity=ip_entity)
            if ip_res_list:
                ip_entity.Location = ip_res_list[0].Location  # type: ignore
        yield ip_entity


@export
def get_map_center(entities: Iterable[Entity], mode: str = "modal"):
    """
    Calculate median point between Entity IP locations.

    Parameters
    ----------
    entities : Iterable[Entity]
        An iterable of entities containing IpAddress geolocation information.
        The entities can be IpAddress entities or other entities that
        have IpAddress properties.
        The entities must all be of the same type.
    mode : str, optional
        The averaging method to use, by default "median".
        "median" and "mean" are the supported values.

    Returns
    -------
    Tuple
        The Latitude and Longitude calculated

    Notes
    -----
    The function uses the first entity in the `entities` to determine
    how to process the collection. E.g. if the first entity has properties
    src_ip and dest_ip of type `IpAddress`, these are the only properties
    that will be processed for the remainder of the entities.

    """
    ip_entities: List[IpAddress] = []
    loc_entities: List[GeoLocation] = []
    if not entities:
        return (0, 0)
    entities = list(entities)
    if isinstance(entities[0], IpAddress):
        return get_center_ip_entities(entities)  # type: ignore
    loc_props = [
        p_name
        for p_name, p_val in entities[0].properties.items()
        if isinstance(p_val, (IpAddress, GeoLocation))
    ]
    for entity, prop in itertools.product(entities, loc_props):
        if prop not in entity:
            continue
        loc_entity = entity[prop]
        if isinstance(loc_entity, IpAddress):
            ip_entities.append(loc_entity)
        elif isinstance(loc_entity, GeoLocation):
            loc_entities.append(loc_entity)
    locs_ips = _extract_locs_ip_entities(ip_entities)
    return get_center_geo_locs(locs_ips + loc_entities, mode=mode)


def _extract_locs_ip_entities(ip_entities: Iterable[IpAddress]):
    """Return the list of IP entities that have a Location property."""
    if isinstance(ip_entities[0], list):  # type: ignore
        return [
            ip[0]["Location"]  # type: ignore
            for ip in ip_entities
            if bool(ip[0].Location)  # type: ignore
        ]
    return [ip["Location"] for ip in ip_entities if bool(ip.Location)]


@export
def get_center_ip_entities(
    ip_entities: Iterable[IpAddress], mode: str = "median"
) -> Tuple[float, float]:
    """
    Return the geographical center of the IP address locations.

    Parameters
    ----------
    ip_entities : Iterable[IpAddress]
        IpAddress entities with location information
    mode : str, optional
        The averaging method to us, by default "median".
        "median" and "mean" are the supported values.

    Returns
    -------
    Tuple[Union[int, float], Union[int, float]]
        Tuple of latitude, longitude

    """
    ip_locs_longs = _extract_locs_ip_entities(ip_entities)
    return get_center_geo_locs(ip_locs_longs, mode=mode)


def _extract_coords_loc_entities(loc_entities: Iterable[GeoLocation]):
    """Return list of coordinate tuples from GeoLocation entities."""
    return [
        (loc["Latitude"], loc["Longitude"])
        for loc in loc_entities
        if "Latitude" in loc and "Longitude" in loc
    ]


@export
def get_center_geo_locs(
    loc_entities: Iterable[GeoLocation], mode: str = "median"
) -> Tuple[float, float]:
    """
    Return the geographical center of the geo locations.

    Parameters
    ----------
    loc_entities : Iterable[GeoLocation]
        GeoLocation entities with location information
    mode : str, optional
        The averaging method to use, by default "median".
        "median" and "mean" are the supported values.

    Returns
    -------
    Tuple[Union[int, float], Union[int, float]]
        Tuple of latitude, longitude

    """
    lat_longs = _extract_coords_loc_entities(loc_entities)
    return _get_center_coords(lat_longs, mode=mode)


def _get_center_coords(
    locations: Iterable[Tuple[float, float]], mode: str = "median"
) -> Tuple[float, float]:
    """Return the center (median) of the coordinates."""
    if not locations:
        return 0, 0
    locs = list(locations)
    if mode == "median":
        with contextlib.suppress(stats.StatisticsError):
            return (
                stats.median([loc[0] for loc in locs if not math.isnan(loc[0])]),
                stats.median([loc[1] for loc in locs if not math.isnan(loc[1])]),
            )
    return (
        stats.mean([loc[0] for loc in locs if not math.isnan(loc[0])]),
        stats.mean([loc[1] for loc in locs if not math.isnan(loc[1])]),
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Azure KeyVault pre-authentication."""
from __future__ import annotations

import logging
import os
import sys
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Callable, ClassVar

from azure.common.credentials import get_cli_profile
from azure.core.credentials import TokenCredential
from azure.identity import (
    AzureCliCredential,
    AzurePowerShellCredential,
    CertificateCredential,
    ChainedTokenCredential,
    ClientSecretCredential,
    DeviceCodeCredential,
    EnvironmentCredential,
    InteractiveBrowserCredential,
    ManagedIdentityCredential,
    VisualStudioCodeCredential,
)
from dateutil import parser

from .._version import VERSION
from ..common.exceptions import MsticpyAzureConfigError

# pylint: disable=unused-import
from .cloud_mappings import AzureCloudConfig, default_auth_methods  # noqa: F401
from .cred_wrapper import CredentialWrapper

__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)


_HELP_URI = (
    "https://msticpy.readthedocs.io/en/latest/getting_started/AzureAuthentication.html"
)


@dataclass
class AzCredentials:
    """Class holding legacy(ADAL) and modern(MSAL) credentials."""

    legacy: TokenCredential
    modern: ChainedTokenCredential


# pylint: disable=too-few-public-methods
class AzureCredEnvNames:
    """Enumeration of Azure environment credential names."""

    AZURE_CLIENT_ID: ClassVar[str] = (
        "AZURE_CLIENT_ID"  # The app ID for the service principal
    )
    AZURE_TENANT_ID: ClassVar[str] = (
        "AZURE_TENANT_ID"  # The service principal's Azure AD tenant ID
    )
    # [SuppressMessage(
    #   "Microsoft.Security",
    #   "CS002:SecretInNextLine",
    #   Justification="This is an enum of env variable names"
    # )]
    AZURE_CLIENT_SECRET: ClassVar[str] = "AZURE_CLIENT_SECRET"  # nosec  # noqa

    # Certificate auth:
    # A path to certificate and private key pair in PEM or PFX format
    AZURE_CLIENT_CERTIFICATE_PATH: ClassVar[str] = "AZURE_CLIENT_CERTIFICATE_PATH"
    # (Optional) The password protecting the certificate file
    # (for PFX (PKCS12) certificates).
    AZURE_CLIENT_CERTIFICATE_PASSWORD: ClassVar[str] = (
        "AZURE_CLIENT_CERTIFICATE_PASSWORD"  # nosec  # noqa
    )
    # (Optional) Specifies whether an authentication request will include an x5c
    # header to support subject name / issuer based authentication.
    # When set to `true` or `1`, authentication requests include the x5c header.
    AZURE_CLIENT_SEND_CERTIFICATE_CHAIN: ClassVar[str] = (
        "AZURE_CLIENT_SEND_CERTIFICATE_CHAIN"
    )

    # Username and password:
    AZURE_USERNAME: ClassVar[str] = (
        "AZURE_USERNAME"  # The username/upn of an AAD user account.
    )
    # [SuppressMessage(
    #   "Microsoft.Security",
    #   "CS002:SecretInNextLine",
    #   Justification="This is an enum of env variable names"
    # )]
    AZURE_PASSWORD: ClassVar[str] = "AZURE_PASSWORD"  # User password  # nosec  # noqa


_VALID_ENV_VAR_COMBOS: tuple[tuple[str, ...], ...] = (
    (
        AzureCredEnvNames.AZURE_CLIENT_ID,
        AzureCredEnvNames.AZURE_CLIENT_SECRET,
        AzureCredEnvNames.AZURE_TENANT_ID,
    ),
    (
        AzureCredEnvNames.AZURE_CLIENT_ID,
        AzureCredEnvNames.AZURE_TENANT_ID,
        AzureCredEnvNames.AZURE_CLIENT_CERTIFICATE_PATH,
    ),
    (
        AzureCredEnvNames.AZURE_CLIENT_ID,
        AzureCredEnvNames.AZURE_USERNAME,
        AzureCredEnvNames.AZURE_PASSWORD,
        AzureCredEnvNames.AZURE_TENANT_ID,
    ),
)


def _build_env_client(
    aad_uri: str | None = None,
    **kwargs,
) -> EnvironmentCredential | None:
    """Build a credential from environment variables."""
    del kwargs
    for env_vars in _VALID_ENV_VAR_COMBOS:
        if all(var in os.environ for var in env_vars):
            return EnvironmentCredential(authority=aad_uri)

    # avoid creating env credential if require envs not set.
    logger.info("'env' credential requested but required env vars not set")
    return None


def _build_cli_client(**kwargs) -> AzureCliCredential:
    """Build a credential from Azure CLI."""
    del kwargs
    return AzureCliCredential()


def _build_msi_client(
    tenant_id: str | None = None,
    aad_uri: str | None = None,
    **kwargs,
) -> ManagedIdentityCredential:
    """Build a credential from Managed Identity."""
    msi_kwargs = kwargs.copy()
    if AzureCredEnvNames.AZURE_CLIENT_ID in os.environ:
        msi_kwargs["client_id"] = os.environ[AzureCredEnvNames.AZURE_CLIENT_ID]

    return ManagedIdentityCredential(
        tenant_id=tenant_id,
        authority=aad_uri,
        **msi_kwargs,
    )


def _build_vscode_client(
    tenant_id: str | None = None,
    aad_uri: str | None = None,
    **kwargs,
) -> VisualStudioCodeCredential:
    """Build a credential from Visual Studio Code."""
    del kwargs
    return VisualStudioCodeCredential(authority=aad_uri, tenant_id=tenant_id)


def _build_interactive_client(
    tenant_id: str | None = None,
    aad_uri: str | None = None,
    **kwargs,
) -> InteractiveBrowserCredential:
    """Build a credential from Interactive Browser logon."""
    return InteractiveBrowserCredential(
        authority=aad_uri,
        tenant_id=tenant_id,
        **kwargs,
    )


def _build_device_code_client(
    tenant_id: str | None = None,
    aad_uri: str | None = None,
    **kwargs,
) -> DeviceCodeCredential:
    """Build a credential from Device Code."""
    return DeviceCodeCredential(authority=aad_uri, tenant_id=tenant_id, **kwargs)


def _build_client_secret_client(
    tenant_id: str | None = None,
    aad_uri: str | None = None,
    **kwargs,
) -> ClientSecretCredential | None:
    """Build a credential from Client Secret."""
    client_id = kwargs.pop("client_id", None)
    client_secret = kwargs.pop("client_secret", None)
    if not client_secret or not client_id:
        logger.info("'clientsecret' credential requested but no params supplied")
        return None
    return ClientSecretCredential(
        authority=aad_uri,
        tenant_id=tenant_id,  # type: ignore
        client_id=client_id,
        client_secret=client_secret,
        **kwargs,
    )


def _build_certificate_client(
    tenant_id: str | None = None,
    aad_uri: str | None = None,
    **kwargs,
) -> CertificateCredential | None:
    """Build a credential from Certificate."""
    client_id = kwargs.pop("client_id", None)
    if not client_id:
        logger.info(
            "'certificate' credential requested but client_id param not supplied"
        )
        return None
    return CertificateCredential(
        authority=aad_uri,
        tenant_id=tenant_id,  # type: ignore
        client_id=client_id,
        **kwargs,  # type: ignore
    )


def _build_powershell_client(**kwargs) -> AzurePowerShellCredential:
    """Build a credential from PowerShell logon session."""
    del kwargs
    return AzurePowerShellCredential()


_CLIENTS: dict[str, Callable] = dict(
    {
        "env": _build_env_client,
        "cli": _build_cli_client,
        "msi": _build_msi_client,
        "vscode": _build_vscode_client,
        "powershell": _build_powershell_client,
        "interactive": _build_interactive_client,
        "interactive_browser": _build_interactive_client,
        "devicecode": _build_device_code_client,
        "device_code": _build_device_code_client,
        "device": _build_device_code_client,
        "environment": _build_env_client,
        "managedidentity": _build_msi_client,
        "managed_identity": _build_msi_client,
        "clientsecret": _build_client_secret_client,
        "client_secret": _build_client_secret_client,
        "certificate": _build_certificate_client,
        "cert": _build_certificate_client,
    }
)


def list_auth_methods() -> list[str]:
    """Return list of accepted authentication methods."""
    return sorted(_CLIENTS.keys())


def _az_connect_core(
    auth_methods: list[str] | None = None,
    cloud: str | None = None,
    tenant_id: str | None = None,
    silent: bool = False,
    *,
    region: str | None = None,
    credential: AzCredentials | None = None,
    **kwargs,
) -> AzCredentials:
    """
    Authenticate using multiple authentication sources.

    Parameters
    ----------
    auth_methods : list[str], optional
        List of authentication methods to try
        For a list of possible authentication methods use the `list_auth_methods`
        function.
        If not set, it will use the value defined in msticpyconfig.yaml.
        If this is not set, the default is ["env", "cli", "msi", "interactive"]
    cloud : str, optional
        What Azure cloud to connect to.
        By default it will attempt to use the cloud setting from config file.
        If this is not set it will default to Azure Public Cloud
    tenant_id : str, optional
        The tenant to authenticate against. If not supplied,
        the tenant ID is read from configuration, or the default tenant for the identity.
    silent : bool, optional
        Whether to display any output during auth process. Default is False.
    credential : AzureCredential
        If an Azure credential is passed, it will be used directly.

    Returns
    -------
    AzCredentials
                Named tuple of:
        - legacy (ADAL) credentials
        - modern (MSAL) credentials

    Raises
    ------
    MsticpyAzureConfigError
        If chained token credential creation fails.

    See Also
    --------
    list_auth_methods

    Notes
    -----
    The function tries to obtain credentials from the following
    sources:
    - Azure Auth Environment variables
    - Azure CLI (if an active session is logged on)
    - Managed Service Identity
    - Interactive browser logon
    - and others - see list_auth_methods for full list.
    If the authentication is successful both ADAL (legacy) and
    MSAL (modern) credential types are returned.

    """
    # Create the auth methods with the specified cloud region
    cloud = cloud or region or AzureCloudConfig().cloud
    az_config = AzureCloudConfig(cloud)
    aad_uri = az_config.authority_uri
    logger.info("az_connect_core - using %s cloud and endpoint: %s", cloud, aad_uri)

    tenant_id = tenant_id or az_config.tenant_id
    auth_methods = auth_methods or az_config.auth_methods
    logger.info(
        "TenantId:  %s, requested auth methods: %s",
        tenant_id,
        ", ".join(auth_methods or ["none"]),
    )

    # Filter and replace error message when credentials not found
    azure_identity_logger: logging.Logger = logging.getLogger("azure.identity")
    handler = logging.StreamHandler(sys.stdout)
    if silent:
        handler.addFilter(_filter_all_warnings)
    else:
        handler.addFilter(_filter_credential_warning)
    azure_identity_logger.setLevel(logging.WARNING)
    azure_identity_logger.handlers = [handler]

    if not credential:
        chained_credential: ChainedTokenCredential = _create_chained_credential(
            aad_uri=aad_uri,
            requested_clients=auth_methods,
            tenant_id=tenant_id,
            **kwargs,
        )
        wrapped_credentials: CredentialWrapper = CredentialWrapper(
            chained_credential, resource_id=az_config.token_uri
        )
        return AzCredentials(wrapped_credentials, chained_credential)  # type: ignore[arg-type]

    # Create the wrapped credential using the passed credential
    wrapped_credentials = CredentialWrapper(credential, resource_id=az_config.token_uri)
    return AzCredentials(
        wrapped_credentials, ChainedTokenCredential(credential)  # type: ignore[arg-type]
    )


az_connect_core: Callable[..., AzCredentials] = _az_connect_core


def _create_chained_credential(
    aad_uri,
    requested_clients: list[str] | None = None,
    tenant_id: str | None = None,
    **kwargs,
) -> ChainedTokenCredential:
    """
    Build a chained token credential.

    Parameters
    ----------
    requested_clients : list[str]
        List of clients to chain.
    aad_uri : str
        The URI of the Azure AD cloud to connect to
    tenant_id : str
        The tenant ID to connect to

    Returns
    -------
    ChainedTokenCredential
        A chained token credential.

    Raises
    ------
    MsticpyAzureConfigError
        If the chained credential creation fails.

    """
    # Create the chained credential
    if not requested_clients:
        requested_clients = ["env", "cli", "msi", "interactive"]
        logger.info("No auth methods requested defaulting to: %s", requested_clients)
    cred_list = []
    invalid_cred_types: list[str] = []
    unusable_cred_type: list[str] = []
    for cred_type in requested_clients:  # type: ignore[union-attr]
        if cred_type not in _CLIENTS:
            invalid_cred_types.append(cred_type)
            logger.info("Unknown authentication type requested: %s", cred_type)
            continue
        cred_client = _CLIENTS[cred_type](
            tenant_id=tenant_id,
            aad_uri=aad_uri,
            **kwargs,
        )
        if cred_client is not None:
            cred_list.append(cred_client)
        else:
            unusable_cred_type.append(cred_type)
    logger.info(
        "Cred types added to chained credential: %s",
        ", ".join(cred.__class__.__name__ for cred in cred_list if cred is not None),
    )
    if not cred_list:
        exception_args = [
            "Cannot authenticate - no valid credential types.",
            "At least one valid authentication method required.",
            f"Configured auth_types: {','.join(requested_clients)}",
        ]
        if invalid_cred_types:
            exception_args.append(
                f"Unrecognized auth_types: {','.join(invalid_cred_types)}",
            )
        if unusable_cred_type:
            exception_args.extend(
                [
                    "The following auth types could not be used due to",
                    "missing parameters or environment variables:",
                    ",".join(invalid_cred_types),
                ]
            )
        raise MsticpyAzureConfigError(
            *exception_args,
            help_uri=_HELP_URI,
            title="Authentication failure",
        )
    return ChainedTokenCredential(*cred_list)


def only_interactive_cred(chained_cred: ChainedTokenCredential):
    """Return True if only interactive browser credentials available."""
    return len(chained_cred.credentials) == 1 and isinstance(
        chained_cred.credentials[0], InteractiveBrowserCredential
    )


def _filter_credential_warning(record) -> bool:
    """Rewrite out credential not found message."""
    if (
        not record.name.startswith("azure.identity")
        or record.levelno != logging.WARNING
    ):
        return True
    message = record.getMessage()
    if ".get_token" in message:
        if message.startswith("EnvironmentCredential"):
            print("Unable to sign-in with environment variable credentials.")
        if message.startswith("AzureCliCredential"):
            print("Unable to sign-in with Azure CLI credentials.")
        if message.startswith("ManagedIdentityCredential"):
            print("Unable to sign-in with Managed Instance credentials.")
    return not message


def _filter_all_warnings(record) -> bool:
    """Filter out credential error messages."""
    if record.name.startswith("azure.identity") and record.levelno == logging.WARNING:
        message = record.getMessage()
        if ".get_token" in message:
            return not message
    return True


class AzureCliStatus(Enum):
    """Enumeration for _check_cli_credentials return values."""

    CLI_OK = 0
    CLI_NOT_INSTALLED = 1
    CLI_NEEDS_SIGN_IN = 2
    CLI_TOKEN_EXPIRED = 3
    CLI_UNKNOWN_ERROR = 4


def check_cli_credentials() -> tuple[AzureCliStatus, str | None]:
    """Check to see if there is a CLI session with a valid AAD token."""
    try:
        cli_profile = get_cli_profile()
        raw_token = cli_profile.get_raw_token()
        bearer_token = None
        if (
            isinstance(raw_token, tuple)
            and len(raw_token) == 3
            and len(raw_token[0]) == 3
        ):
            bearer_token = raw_token[0][2]
            if (
                parser.parse(bearer_token.get("expiresOn", datetime.min))
                < datetime.now()
            ):
                raise ValueError("AADSTS70043: The refresh token has expired")

        return AzureCliStatus.CLI_OK, "Azure CLI credentials available."
    except ImportError:
        # Azure CLI not installed
        return AzureCliStatus.CLI_NOT_INSTALLED, None
    except Exception as ex:  # pylint: disable=broad-except
        if "AADSTS70043: The refresh token has expired" in str(ex):
            message = (
                "Azure CLI was detected but the token has expired. "
                "For Azure CLI single sign-on, please sign in using '!az login'."
            )
            return AzureCliStatus.CLI_TOKEN_EXPIRED, message
        if "Please run 'az login' to setup account" in str(ex):
            message = (
                "Azure CLI was detected but no token is available. "
                "For Azure CLI single sign-on, please sign in using '!az login'."
            )
            return AzureCliStatus.CLI_NEEDS_SIGN_IN, message
        return AzureCliStatus.CLI_UNKNOWN_ERROR, None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""VirusTotal File Behavior functions."""
from __future__ import annotations

import logging
import re
from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from pprint import pformat
from typing import TYPE_CHECKING, Any, ClassVar

import attr
import ipywidgets as widgets
import numpy as np
import pandas as pd
from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyImportExtraError, MsticpyUserError
from ...transform.proc_tree_builder import ProcSchema, build_proc_tree

if TYPE_CHECKING:
    from bokeh.plotting import figure

try:
    import vt
except ImportError as imp_err:
    err_msg: str = (
        "Cannot use this feature without vt-py and vt-graph-api packages installed."
    )
    raise MsticpyImportExtraError(
        err_msg,
        title="Error importing VirusTotal modules.",
        extra="vt3",
    ) from imp_err
logger: logging.Logger = logging.getLogger(__name__)
__version__ = VERSION
__author__ = "Ian Hellen"


VT_API_NOT_FOUND = "NotFoundError"


_FB_CAT_PATTERNS: dict[str, str] = {
    "File": "file.*",
    "Process": "process.*|command.*|module.*",
    "Registry": "registry.*",
    "Network": ".*ips|dns.*|.*urls|ip.*|http.*|tls",
    "System": "mutex.*|calls.*|permissions.*|text.*",
    "Other": ".*",
}

_BORDER_LAYOUT = widgets.Layout(
    width="90%",
    border="solid gray 1px",
    margin="1pt",
    padding="5pt",
)


class VTFileBehavior:
    """VirusTotal File Behavior class."""

    _SANDBOXES: ClassVar[list[str]] = [
        "f_secure_sandbox",
        "bitdam_atp",
        "vmray",
        "virustotal_zenbox",
        "sangfor_zsand",
        "virustotal_jujubox",
        "qianxin_reddrip",
        "nsfocus_poma",
        "virustotal_androbox",
        "venuseye_sandbox",
        "cyber_adapt",
        "dr_web_vxcube",
        "virustotal_observer",
        "tencent_habo",
        "yomi_hunter",
        "virustotal_jsbox",
        "virustotal_cuckoofork",
        "lastline",
        "reaqta_hive",
        "os_x_sandbox",
        "virustotal_droidy",
        "sndbox",
        "virustotal_r2dbox",
        "rising_moves",
        "virustotal_box_of_apples",
        "secondwrite",
        "malwation",
        "c2ae",
        "microsoft_sysinternals",
    ]

    _FP_ENDPOINTS: ClassVar[dict[str, str]] = {
        "summary": "/files/{id}/behaviour_summary",
        "sandbox": "/file_behaviours/{id}_{sandbox}",
        "evtx": "/file_behaviours/{sandbox_id}/evtx",
        "pcap": "/file_behaviours/{sandbox_id}/pcap",
        "memdump": "/file_behaviours/{sandbox_id}/memdump",
    }

    @classmethod
    def list_sandboxes(cls: type[Self]) -> list[str]:
        """Return list of known sandbox types."""
        return list(cls._SANDBOXES)

    def __init__(
        self: Self,
        vt_key: str,
        file_id: str | None = None,
        file_summary: pd.DataFrame | pd.Series | dict[str, Any] | None = None,
    ) -> None:
        """
        Initialize the VTFileBehavior class.

        Parameters
        ----------
        vt_key : str, optional
            VirusTotal API key, by default None
        file_id : Optional[str], optional
            The ID of the file to look up, by default None
        file_summary : Optional[Union[pd.DataFrame, pd, Series, dict[str, Any]]], optional
            VT file summary - this can be in one of the following formats:
            VT object dictionary
            Pandas DataFrame - first row is assumed to be the file summary
            Pandas Series
            by default None

        """
        self._vt_client = vt.Client(apikey=vt_key)
        if file_id is None and file_summary is None:
            error_msg: str = "You must supply either a file_id or a file_summary."
            raise MsticpyUserError(
                error_msg,
                title="Missing required parameter.",
            )

        if isinstance(file_summary, pd.DataFrame):
            file_summary = file_summary.iloc[0]
        if isinstance(file_summary, pd.Series):
            file_summary = file_summary.to_dict()
        self.file_summary: dict[str, Any] = file_summary or {}
        self.file_id: str | Any | None = file_id or self.file_summary.get("id")

        self._file_behavior: dict[str, Any] = {}
        self.categories: dict[str, Any] = {}
        self.behavior_links: dict[str, Any] = {}
        self.process_tree_df: pd.DataFrame | None = None

    def _reset_summary(self: Self) -> None:
        self._file_behavior = {}
        self.categories = {}
        self.process_tree_df = None

    @property
    def sandbox_id(self: Self) -> str:
        """Return sandbox ID of detonation."""
        return self.categories.get("id", "")

    @property
    def has_evtx(self: Self) -> bool:
        """Return True if EVTX data is available (Enterprise only)."""
        return self.categories.get("has_evtx", False)

    @property
    def has_memdump(self: Self) -> bool:
        """Return True if memory dump data is available (Enterprise only)."""
        return self.categories.get("has_memdump", False)

    @property
    def has_pcap(self: Self) -> bool:
        """Return True if PCAP data is available (Enterprise only)."""
        return self.categories.get("has_pcap", False)

    def get_file_behavior(self: Self, sandbox: str | None = None) -> None:
        """
        Retrieve the file behavior data.

        Parameters
        ----------
        sandbox : str, optional
            Name of specific sandbox to retrieve, by default None
            If None, it will retrieve the behavior summary.

        """
        if sandbox:
            endpoint: str = self._FP_ENDPOINTS["sandbox"].format(
                id=self.file_id,
                sandbox=sandbox,
            )
        else:
            endpoint = self._FP_ENDPOINTS["summary"].format(id=self.file_id)

        try:
            self._file_behavior = self._vt_client.get_data(endpoint)
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                self._file_behavior = {"id": self.file_id, "result": VT_API_NOT_FOUND}
                return
            raise
        finally:
            self._vt_client.close()

        if "attributes" in self._file_behavior:
            self.categories = self._file_behavior.get("attributes", {})
            self.behavior_links = self._file_behavior.get("links", {})
        else:
            self.categories = self._file_behavior

    def browse(self: Self) -> widgets.VBox | None:
        """Browse the behavior categories."""
        if not self.has_behavior_data:
            self._print_no_data()
            return None
        groupings: dict[str, Any] = {}
        remaining_categories: set[str] = set(self.categories)
        for name, pattern in _FB_CAT_PATTERNS.items():
            groupings[name] = _extract_subcats(pattern, remaining_categories)
            remaining_categories = remaining_categories - groupings[name]

        accordion = widgets.Accordion()
        child_tabs: dict[str, Any] = {}
        for group, sub_cats in groupings.items():
            sub_cat_tab = widgets.Tab()
            tab_content: dict[str, widgets.HTML] = {
                section: widgets.HTML(value=_format_widget_data(items))
                for section, items in self.categories.items()
                if items and section in sub_cats
            }
            sub_cat_tab.children = list(tab_content.values())
            for idx, section in enumerate(tab_content):
                sub_cat_tab.set_title(idx, section)
            child_tabs[group] = sub_cat_tab

        accordion.children = list(child_tabs.values())
        for idx, group_name in enumerate(child_tabs):
            accordion.set_title(idx, group_name)
        accordion.selected_index = 0

        html_title = widgets.HTML(
            ("<h2>VirusTotal Detonation Details</h2> For file {self.file_id}"),
            layout=_BORDER_LAYOUT,
        )
        return widgets.VBox([html_title, accordion])

    @property
    def process_tree(self: Self) -> figure | None:
        """Return the process tree plot."""
        if not self.has_behavior_data:
            self._print_no_data()
            return None
        if self.process_tree_df is None:
            self.process_tree_df = _build_process_tree(self.categories)
        plot, _ = self.process_tree_df.mp_plot.process_tree(
            schema=VT_PROCSCHEMA,
            legend_col="name",
            hide_legend=True,
        )
        return plot

    @property
    def has_behavior_data(self: Self) -> bool:
        """Return true if file behavior data available."""
        return bool(self.categories)

    def _print_no_data(self: Self) -> None:
        """Print a message if operation is tried with no data."""
        logger.info("No data available for %s.", self.file_id)


# Process tree extraction


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True)
class SIProcess:
    """Data class to hold each process from detonation."""

    process_id: str
    name: str
    cmd_line: str
    parent_id: int = -1
    proc_key: str | None = None
    parent_key: str | None = None
    path: str | None = None
    IsRoot: bool = False
    IsLeaf: bool = False
    IsBranch: bool = False
    children: list = attr.ib(factory=list)
    time_offset: int = 0


# pylint: enable=too-few-public-methods


VT_PROCSCHEMA = ProcSchema(
    process_name="name",
    process_id="process_id",
    parent_id="parent_id",
    cmd_line="cmd_line",
    time_stamp="time_stamp",
    logon_id="logon_id",
    path_separator="\\",
    user_name="user_name",
    host_name_column="host",
    event_id_column="event_id",
)


def _build_process_tree(fb_categories: dict[str, Any]) -> pd.DataFrame:
    """Top level function to create displayable DataFrame."""
    proc_tree_raw: list[dict[str, Any]] = deepcopy(fb_categories["processes_tree"])
    procs_created: dict[str, Any] = {
        Path(proc).parts[-1].lower(): proc
        for proc in fb_categories["processes_created"]
    }

    si_procs: list[SIProcess] = _extract_processes(proc_tree_raw, procs_created)
    process_tree_df: pd.DataFrame = pd.DataFrame(_procs_to_df(si_procs)).drop(
        columns="children",
    )
    process_tree_df = _try_match_commandlines(
        fb_categories["command_executions"],
        process_tree_df,
    )
    return _fill_missing_proc_tree_values(process_tree_df)


def _extract_processes(
    process_data: list[dict[str, Any]],
    procs_created: dict[str, Any],
    parent: SIProcess | None = None,
) -> list[SIProcess]:
    """Convert processes_tree attribute to SIProcessObjects."""
    procs: list[SIProcess] = []
    for process in process_data:
        si_proc: SIProcess = _create_si_proc(process, procs_created)
        # pylint: disable=invalid-name
        if parent:
            si_proc.parent_key = parent.proc_key
            si_proc.IsBranch = True
        else:
            si_proc.IsRoot = True
        child_procs_raw: list[dict[str, Any]] = process.get("children", [])
        if child_procs_raw:
            si_proc.children = _extract_processes(
                child_procs_raw,
                procs_created,
                parent=si_proc,
            )
        else:
            si_proc.IsLeaf = True
            si_proc.IsBranch = False
        procs.append(si_proc)
    return procs


def _create_si_proc(
    raw_proc: dict[str, Any],
    procs_created: dict[str, Any],
) -> SIProcess:
    """Return an SIProcess Object from a raw VT proc definition."""
    name: str = raw_proc["name"]
    raw_proc["cmd_line"] = name
    for proc in procs_created:
        if name.lower().endswith(proc):
            raw_proc["name"] = procs_created[proc]
            break
    raw_proc["proc_key"] = raw_proc["process_id"] + "|" + raw_proc["name"]
    return SIProcess(**raw_proc)


# Convert to DF
def _procs_to_df(procs: list[SIProcess]) -> list[dict[str, Any]]:
    """Convert the SIProcess objects recursively to a list."""
    df_list: list[dict[str, Any]] = []
    for proc in procs:
        df_list.append(attr.asdict(proc))
        if proc.children:
            df_list.extend(_procs_to_df(proc.children))
    return df_list


# Try to Match up 'command_executions' commandline data with
# process_df.
def _try_match_commandlines(
    command_executions: list,
    procs_cmds: pd.DataFrame,
) -> pd.DataFrame:
    """Return DF with matched commandlines."""
    procs_cmd: pd.DataFrame = procs_cmds.copy()
    procs_cmd["cmd_line"] = np.nan
    procs_cmd["cmd_line"] = procs_cmd["cmd_line"].astype(object)  # Set dtype to object
    weak_matches = 0
    for cmd in command_executions:
        for idx, row in procs_cmd.iterrows():
            if (
                not isinstance(row["cmd_line"], str)
                and np.isnan(row["cmd_line"])
                and row["name"] in cmd
            ):
                procs_cmd.loc[idx, "cmd_line"] = cmd  # type: ignore[reportCallIssue]
                break
    for cmd in command_executions:
        for idx, row in procs_cmd.iterrows():
            if (
                not isinstance(row["cmd_line"], str)
                and np.isnan(row["cmd_line"])
                and Path(row["name"]).stem.lower() in cmd.lower()
            ):
                weak_matches += 1
                procs_cmd.loc[idx, "cmd_line"] = cmd  # type: ignore[reportCallIssue]
                break

    if weak_matches:
        logger.warning(
            "%s of the %d commandlines were weakly matched - some commandlines may be attributed"
            "to the wrong instance of the process.",
            weak_matches,
            len(command_executions),
        )
    return procs_cmd


def _fill_missing_proc_tree_values(process_df: pd.DataFrame) -> pd.DataFrame:
    # Define a schema to map Df names on to internal ProcSchema
    process_df["path"] = np.nan
    process_df["path"] = process_df["path"].astype(object)  # Set dtype to object
    process_df.loc[process_df.IsRoot, "path"] = pd.Series(
        process_df[process_df.IsRoot].index.astype("str"),
    ).apply(lambda x: x.zfill(5))

    # Fill in some required fields with placeholder data
    process_df["time_stamp"] = datetime.now(tz=timezone.utc)
    process_df["host"] = "sandbox"
    process_df["logon_id"] = "na"
    process_df["event_id"] = "na"
    process_df["source_index"] = pd.Series(process_df.index.astype("str")).apply(
        lambda x: x.zfill(5),
    )

    proc_tree: pd.DataFrame = process_df.set_index("proc_key")

    first_unique: np.ndarray[Any, np.dtype[np.bool_]] = proc_tree.index.duplicated()
    proc_tree = proc_tree[~first_unique]
    # msticpy function to build the tree
    return build_proc_tree(proc_tree)


# Process browser helper functions
def _extract_subcats(pattern: str, categs: set[str]) -> set:
    """Extract the category names matching `pattern`."""
    return {cat for cat in categs if re.match(pattern, cat)}


def _format_widget_data(data_item: list[dict[str, Any] | str] | None) -> str:
    if not data_item:
        return ""
    if isinstance(data_item, list):
        if isinstance(data_item[0], dict):
            return pd.DataFrame(data_item).style.hide(axis="index").to_html()
        if isinstance(data_item[0], str):
            return pd.DataFrame(pd.Series(data_item)).style.hide(axis="index").to_html()
    return f"<pre>{pformat(data_item)}</pre>"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
MSTICPy sectools.

.. warning: This sub-package is deprecated.
   All functionality has been removed from this sub-package and moved
   to other sub-packages:

- TI providers -> msticpy.context.tiproviders
  (including vtlookup and vtlookupv3)
- auditdextract -> msticpy.transform
- base64unpack  -> msticpy.transform
- cmd_line -> msticpy.context
- domain_utils -> msticpy.context
- eventcluster -> msticpy.analysis
- geoip -> msticpy.context
- iocextract -> msticpy.transform
- ip_utils -> msticpy.context
- proc_tree_builder -> msticpy.transform
- proc_tree_build_mde -> msticpy.transform
- proc_tree_build_winlx -> msticpy.transform
- proc_tree_schema -> msticpy.transform
- proc_tree_utils -> msticpy.transform
- sectools_magics -> msticpy.init.nbmagics
- syslog_utils -> msticpy.analysis

The sectools sub-package will be removed in version 2.0.0

"""
from .._version import VERSION
from ..lazy_importer import lazy_import

__version__ = VERSION

_LAZY_IMPORTS = {
    "msticpy.context.geoip.GeoLiteLookup",
    "msticpy.context.geoip.IPStackLookup",
    "msticpy.context.geoip.geo_distance",
    "msticpy.context.tilookup.TILookup",
    "msticpy.transform.base64unpack as base64",
    "msticpy.transform.iocextract.IoCExtract",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
MSTICPy Analysis Tools.

This sub-package has a few classes of analysis tools:

- data - this sub-package contains data-processing
  tools for dealing with tranforming data (e.g. auditd, process tree)
  or extracting specific data formats (iocextract, b64, cmdline)
- other items are miscellaneous analysis modules such anomalous
  sequence, time series, eventcluster.

"""

from .._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Dataframe viewer using Panel Tabulator."""
from functools import partial
from pprint import pformat
from textwrap import wrap
from typing import Any, Callable, Dict, Iterable, List, Optional

import pandas as pd
from IPython.display import display

from ..common.exceptions import MsticpyMissingDependencyError

try:
    import panel as pn
except ImportError as err:
    raise MsticpyMissingDependencyError(
        "This component needs the panel package.",
        "Install the conda or pip 'panel' package.",
        packages="panel",
    ) from err


# pylint: disable=too-many-instance-attributes
class DataViewer:
    """Data viewer class."""

    _DEF_HEIGHT = 550
    _DEFAULT_HIDDEN_COLS = ["TenantId"]

    def __init__(self, data: pd.DataFrame, selected_cols: List[str] = None, **kwargs):
        """
        Initialize the DataViewer class.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame to view
        selected_cols : List[str], optional
            Initial subset of columns to show, by default None (all cols)

        Other Parameters
        ----------------
        selectable : Union[bool, str], optional
            Whether rows should be selectable, by default "checkbox"
        show_index : bool, optional
            If True show the DataFrame index as a column, by default  True.
        show_tenant_id : bool, optional
            If True show the TenantId column, by default False.
        max_col_width : int, optional
            Sets the maximum column width to display, by default 500
        detail_cols : List[str]
            List of columns for which details are displayed in collapsible
            field beneath each table row.
        kwargs :
            Other keyword arguments are passed to the panel
            Tabulator control.

        Notes
        -----
        Main attributes:
        value - original dataframe
        selected - indexes of currently selected rows
        selected_dataframe - currently selected rows
        current_view - current dataframe after filtering and sorting
        selection - indexes of currently selected rows

        See Also
        --------
        Tabulator - https://panel.holoviz.org/reference/widgets/Tabulator.html

        """
        if not pn.extension._loaded:
            pn.extension(
                "tabulator",
                sizing_mode="stretch_width",
                css_files=[pn.io.resources.CSS_URLS["font-awesome"]],
            )
        if data.empty:
            raise ValueError("No data available in 'data'")

        # Drop empty columns
        self.data = data.dropna(axis="columns", how="all")
        # Set up hidden columns
        hidden_cols = kwargs.pop("hidden_cols", None)
        self._hidden_columns = self._default_hidden_cols(selected_cols, hidden_cols)

        if (
            not kwargs.pop("show_tenant_id", False)
            and "TenantId" in self._hidden_columns
        ):
            self._hidden_columns.remove("TenantId")

        # Create the tabulator control
        self.data_table = pn.widgets.Tabulator(
            self.data,
            header_filters=kwargs.pop("header_filters", True),
            selectable=kwargs.pop("selectable", "checkbox"),
            show_index=kwargs.pop("show_index", False),
            configuration=self._create_configuration(kwargs),
            pagination="local",
            row_content=self._create_row_formatter(kwargs.pop("detail_cols", None)),
            embed_content=False,
            height=kwargs.pop("height", self._DEF_HEIGHT),
            **kwargs,
        )
        # Add the column chooser
        self.column_chooser = DataTableColumnChooser(
            data,
            selected_cols=selected_cols
            or list(set(data.columns) - set(self._hidden_columns)),  # type: ignore
        )
        self.column_chooser.apply_button.on_click(self._update_columns)
        self.accordion = pn.layout.Accordion(
            ("Select columns", self.column_chooser.layout)
        )
        self._update_columns(btn=None)
        # set layout for the widget.
        self.layout = pn.layout.Column(self.data_table, self.accordion)

    def __getattr__(self, attrib: str) -> Any:
        """Return attribute from underlying Tabulator control."""
        if hasattr(self.data_table, attrib):
            return getattr(self.data_table, attrib)
        raise AttributeError(f"Attribute {attrib} not found.")

    def display(self):
        """Display the widget."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()

    def _update_columns(self, btn):
        """Update the displayed columns."""
        del btn
        hidden_cols = set(self.data.columns) - set(self.column_chooser.selected_columns)
        if self._hidden_columns:
            hidden_cols = hidden_cols | set(self._hidden_columns)
        self.data_table.hidden_columns = list(hidden_cols)
        self.accordion.active = []

    def _create_row_formatter(
        self, detail_columns: Optional[List[str]] = None
    ) -> Optional[Callable]:
        """Build formatter function for row-details."""
        if not detail_columns:
            return None
        row_view_cols = set(detail_columns) & set(self.data.columns)
        return partial(_display_column_details, columns=row_view_cols)

    def _create_configuration(self, kwargs) -> Dict[str, Any]:
        """Create Tabulator configuration dict to pass to JS Tabulator."""
        return {
            "columnDefaults": {"maxWidth": kwargs.pop("max_col_width", 500)},
            "clipboardCopyRowRange": "selected",
            "clipboardCopyConfig": {
                "columnHeaders": True,
                "columnGroups": False,
                "rowGroups": False,
                "columnCalcs": False,
                "dataTree": False,
                "formatCells": False,
            },
        }

    def _default_hidden_cols(self, selected_cols, hidden_cols) -> List[str]:
        """Return list of of columns hidden by default."""
        return [
            hidden_col
            for hidden_col in self._DEFAULT_HIDDEN_COLS
            if (
                hidden_col in self.data.columns
                and selected_cols
                and hidden_col not in selected_cols
            )
        ] + (hidden_cols or [])


class DataTableColumnChooser:
    """DataTableColumnChooser class."""

    def __init__(self, data, selected_cols=None):
        """Initialize the DataTableColumnChooser class."""
        self.data = data
        self._all_col_names = list(data.columns)
        self._initial_cols = selected_cols or self._all_col_names
        self._col_select = pn.widgets.MultiSelect(
            name="Select columns to display",
            value=self._initial_cols,
            options=self._all_col_names,
            size=15,
        )
        self.apply_button = pn.widgets.Button(name="Apply columns")
        self.layout = pn.layout.Column(self._col_select, self.apply_button)

    @property
    def selected_columns(self) -> List[str]:
        """Return a list of Bokeh column definitions for the DataFrame."""
        return self._col_select.value

    @property
    def dataframe_columns(self):
        """Return the selected set of DataFrame columns."""
        return self.data[self._reorder_cols(self.selected_columns)]

    def _reorder_cols(self, columns: List[str]) -> List[str]:
        """Return column list in original order."""
        # order the columns as originally specified (or as the DF)
        col_init = [col for col in self._initial_cols if col in columns]
        # If any new columns, add them to the end of the list
        col_init.extend(list(set(columns) - set(col_init)))
        return col_init

    def display(self):
        """Display in IPython."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


# RowFunction display
_ROW_FORMAT_FUNCS = {
    (dict, list): lambda value, width: f"<pre>{pformat(value, width=width)}</pre>",
    str: lambda value, width: "<br>".join(wrap(value, width=width)),
}
# expand
_FORMATTERS = {
    val_type: func
    for val_tuple, func in _ROW_FORMAT_FUNCS.items()
    for val_type in (val_tuple if isinstance(val_tuple, tuple) else (val_tuple,))
}


def _format_value(value: Any, width=150) -> str:
    """Apply format function to column cell."""
    return _FORMATTERS.get(type(value), _FORMATTERS[str])(value, width)


TD_ALIGN = """
<style>
td { text-align:left; vertical-align: top }
</style>
"""


def _display_column_details(
    row: pd.Series, columns: Iterable[str], width: int = 120
) -> pn.pane.HTML:
    """Return HTML pane with formatted row columns."""
    display_text = "".join(
        f"<tr><td><b>{column}</b></td><td>{_format_value(row[column], width=width)}</td></tr>"
        for column in columns
    )
    return pn.pane.HTML(
        f"{TD_ALIGN}<table>{display_text}</table>", sizing_mode="stretch_width"
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Helper functions for configuration settings."""
from __future__ import annotations

from dataclasses import dataclass, field
import os
import warnings
from collections import UserDict
from typing import Any, Callable

from .._version import VERSION
from .exceptions import MsticpyImportExtraError
from .pkg_config import get_config, refresh_config

try:
    from ..auth.secret_settings import SecretsClient

    _SECRETS_ENABLED = True
except ImportError:
    _SECRETS_ENABLED = False


__version__ = VERSION
__author__ = "Ian Hellen"


class ProviderArgs(UserDict):
    """ProviderArgs dictionary."""

    def __getitem__(self, key) -> Any:
        """Return key value via SecretsClient.read_secret."""
        if key not in self.data:
            raise KeyError(key)
        if _SECRETS_ENABLED:
            return SecretsClient.read_secret(self.data[key])
        return self.data[key]


@dataclass
class ProviderSettings:
    """Provider settings."""

    name: str
    description: str
    provider: str | None = field(default=None)
    args: ProviderArgs = field(default_factory=ProviderArgs)
    primary: bool = field(default=False)


def _secrets_enabled() -> bool:
    return _SECRETS_ENABLED and _SECRETS_CLIENT


def get_secrets_client_func() -> Callable[..., "SecretsClient" | None]:
    """
    Return function to get or create secrets client.

    Returns
    -------
    Callable
        Function to get, replace or create a SecretsClient

    Notes
    -----
    This function creates closure that persists the secrets client
    instance.
    The inner function works as follows:

    - if called with no parameters and SecretsClient is not
      instantiated, it will try to instantiate a SecretsClient,
      assign it to the nonlocal `_secrets_client` and return this instance.
    - if called subsequently it will just return the secrets client.
    - if called with a SecretsClient instance as a parameter, it will
      replace the SecretsClient instance and return that.

    """
    _secrets_client: "SecretsClient" | None = None

    def _return_secrets_client(
        secrets_client: "SecretsClient" | None = None, **kwargs
    ) -> "SecretsClient" | None:
        """Return (optionally setting or creating) a SecretsClient."""
        nonlocal _secrets_client
        if not _SECRETS_ENABLED:
            return None
        if isinstance(secrets_client, SecretsClient):
            _secrets_client = secrets_client
        if _secrets_client is None:
            _secrets_client = SecretsClient(**kwargs)
        return _secrets_client

    return _return_secrets_client


# Create a SecretsClient instance if it can be imported when
# the module is imported.
_SECRETS_CLIENT: Any = None
# Create the secrets client closure
_SET_SECRETS_CLIENT: Callable[..., "SecretsClient" | None] = get_secrets_client_func()
# Create secrets client instance if SecretsClient can be imported
# and config has KeyVault settings.
if get_config("KeyVault", None) and _SECRETS_ENABLED:
    _SECRETS_CLIENT = _SET_SECRETS_CLIENT()


def get_provider_settings(config_section="TIProviders") -> dict[str, ProviderSettings]:
    """
    Read Provider settings from package config.

    Parameters
    ----------
    config_section : str, optional
        [description], by default "TIProviders"

    Returns
    -------
    dict[str, ProviderSettings]
        Provider settings indexed by provider name.

    """
    # pylint: disable=global-statement
    global _SECRETS_CLIENT
    # pylint: enable=global-statement
    if get_config("KeyVault", None):
        if _SECRETS_CLIENT is None and _SECRETS_ENABLED:
            print(
                "KeyVault enabled. Secrets access may require additional authentication."
            )
            _SECRETS_CLIENT = _SET_SECRETS_CLIENT()
    else:
        _SECRETS_CLIENT = None
    section_settings = get_config(config_section, None)
    if not section_settings:
        return {}

    settings = {}
    for provider, item_settings in section_settings.items():
        prov_args = item_settings.get("Args")
        prov_settings = ProviderSettings(  # type: ignore[call-arg]
            name=provider,
            description=item_settings.get("Description"),
            args=_get_setting_args(
                config_path=config_section,
                provider_name=provider,
                prov_args=prov_args,
            ),
            primary=item_settings.get("Primary", False),
            provider=item_settings.get("Provider", provider),
        )
        settings[provider] = prov_settings

    return settings


def reload_settings():
    """Reload settings from config files."""
    refresh_config()


def refresh_keyring():
    """Refresh local keyring secrets cache from Key Vault."""
    if _secrets_enabled():
        _SECRETS_CLIENT.refresh_keyring()


def clear_keyring():
    """Delete local keyring secrets cache."""
    if _secrets_enabled():
        _SECRETS_CLIENT.clear_keyring_secrets()


def auth_secrets_client(
    tenant_id: str | None = None,
    auth_methods: list[str] | None = None,
    credential: Any = None,
    **kwargs,
) -> None:
    """
    Authenticate the Secrets/Key Vault client.

    Parameters
    ----------
    auth_methods : List[str], optional
        List of authentication methods to try
        Possible options are:
        - "env" - to get authentication details from environment variables
        - "cli" - to use Azure CLI authentication details
        - "msi" - to user Managed Service Identity details
        - "interactive" - to prompt for interactive login
        - "vscode" - to use VSCode credentials
        - "powershell" - to use PowerShell credentials
        - "interactive" - to prompt for interactive login
        - "cache" - to use shared token cache credentials
        - "devicecode" - to use device code with web login
        - "clientsecret" - to use client id/secret login.
        Default is ["env", "msi", "vscode", "cli", "powershell", "interactive"]
    tenant_id : str, optional
        The tenant to authenticate against. If not supplied, the default
        tenant for the identity will be used.
    silent : bool, optional
        Set True to hide all output during connection, by default False
    credential : Azure Credential, optional
        Use an existing Azure credential to authenticate.

    Other Parameters
    ----------------
    client_id : str
        Required when using "clientsecret" auth method
    client_secret : str
        Required when using "clientsecret" auth method

    """
    if _secrets_enabled():
        secrets_client: SecretsClient = SecretsClient(
            tenant_id=tenant_id,
            auth_methods=auth_methods,
            credential=credential,
            **kwargs,
        )
        _SET_SECRETS_CLIENT(secrets_client=secrets_client)


def get_protected_setting(config_path, setting_name) -> Any:
    """Return a potentially protected setting value."""
    config_settings = get_config(config_path)
    prov_args = _get_protected_settings(config_path, config_settings)
    return prov_args.get(setting_name)


def _get_setting_args(
    config_path: str,
    provider_name: str,
    prov_args: dict[str, Any] | None,
) -> ProviderArgs:
    """Extract the provider args from the settings."""
    if not prov_args:
        return ProviderArgs()
    name_map: dict[str, str] = {
        "workspaceid": "workspace_id",
        "tenantid": "tenant_id",
        "subscriptionid": "subscription_id",
    }
    return _get_protected_settings(
        setting_path=f"{config_path}.{provider_name}.Args",
        section_settings=prov_args,
        name_map=name_map,
    )


def _get_protected_settings(
    setting_path: str,
    section_settings: dict[str, Any] | None,
    name_map: dict[str, str] | None = None,
) -> ProviderArgs:
    """
    Lookup configuration values config, environment or KeyVault.

    Parameters
    ----------
    setting_path : str
        Dotted path to the setting
    section_settings : Optional[dict[str, Any]]
        The configuration settings for this path.
    name_map : Optional[dict[str, str]], optional
        Optional mapping to re-write setting names,
        by default None

    Returns
    -------
    ProviderArgs
        Dictionary of resolved settings

    """
    if not section_settings:
        return ProviderArgs()
    setting_dict: ProviderArgs = ProviderArgs(section_settings.copy())

    for arg_name, arg_value in section_settings.items():
        target_name: str = arg_name
        if name_map:
            target_name = name_map.get(target_name.casefold(), target_name)

        try:
            setting_dict[target_name] = _fetch_secret_setting(
                f"{setting_path}.{arg_name}", arg_value
            )
        except NotImplementedError:
            warnings.warn(f"Setting type for setting {arg_value} not yet implemented. ")
    return setting_dict


def _fetch_secret_setting(
    setting_path: str,
    config_setting: str | dict[str, Any],
) -> str | Callable[[], Any] | None:
    """
    Return required value for potential secret setting.

    Parameters
    ----------
    setting_path : str
        Dotted path to the setting
    config_setting : Union[str, dict[str, Any]]
        Setting value (str or Dict)

    Returns
    -------
    Union[Optional[str], Callable[[], Any]]
        Either a string or accessor function.

    Raises
    ------
    MsticpyImportExtraError
        _description_
    NotImplementedError
        _description_

    """
    if isinstance(config_setting, str):
        return config_setting
    if not isinstance(config_setting, dict):
        err_msg: str = (
            "Configuration setting format not recognized. "
            f"'{setting_path}' should be a string or dictionary "
            "with either 'EnvironmentVar' or 'KeyVault' entry."
        )
        raise NotImplementedError(err_msg)
    if "EnvironmentVar" in config_setting:
        env_value: str | None = os.environ.get(config_setting["EnvironmentVar"])
        if not env_value:
            warnings.warn(
                f"Environment variable {config_setting['EnvironmentVar']}"
                f" ({setting_path})"
                " was not set"
            )
        return env_value
    if "KeyVault" in config_setting:
        if not _SECRETS_ENABLED:
            err_msg = "Cannot use this feature without Key Vault support installed"
            raise MsticpyImportExtraError(
                err_msg,
                title="Error importing Loading Key Vault and/or keyring libraries.",
                extra="keyvault",
            )
        if not _SECRETS_CLIENT:
            warnings.warn(
                "Cannot use a KeyVault configuration setting without"
                "a KeyVault configuration section in msticpyconfig.yaml"
                f" ({setting_path})",
                stacklevel=1,
            )
            return None
        return _SECRETS_CLIENT.get_secret_accessor(setting_path)
    err_msg = (
        "Configuration setting format not recognized. "
        f"'{setting_path}' should be a string or dictionary "
        "with either 'EnvironmentVar' or 'KeyVault' entry."
    )
    raise NotImplementedError(err_msg)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTICPy query template definition."""
from dataclasses import field
from typing import Any, Dict, List, Optional, Union

from pydantic.dataclasses import dataclass

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-instance-attributes
@dataclass
class QueryMetadata:
    """Metadata for query definitions."""

    version: int
    description: str
    data_environments: List[str]
    data_families: List[str]
    database: Optional[str] = None
    cluster: Optional[str] = None
    clusters: Optional[List[str]] = None
    cluster_groups: Optional[List[str]] = None
    tags: List[str] = field(default_factory=list)
    data_source: Optional[str] = None
    aliases: Optional[Union[str, List[str]]] = None
    query_macros: Optional[Dict[str, Any]] = None


@dataclass
class QueryParameter:
    """Query parameter."""

    description: str
    datatype: str
    default: Any = None
    aliases: Optional[List[str]] = None


@dataclass
class QueryDefaults:
    """Default values for query definitions."""

    metadata: Optional[Dict[str, Any]] = None
    parameters: Dict[str, QueryParameter] = field(default_factory=dict)


@dataclass
class QueryArgs:
    """Query arguments."""

    query: str = ""
    uri: Optional[str] = None


@dataclass
class Query:
    """A Query definition."""

    description: str
    args: QueryArgs = field(default_factory=QueryArgs)
    metadata: Optional[Dict[str, Any]] = field(default_factory=dict)  # type: ignore
    parameters: Optional[Dict[str, QueryParameter]] = field(default_factory=dict)  # type: ignore


@dataclass
class QueryCollection:
    """Query Collection class - a query template."""

    metadata: QueryMetadata
    defaults: Optional[QueryDefaults] = None
    sources: Dict[str, Query] = field(default_factory=dict)
    file_name: Optional[str] = None



#  -------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for
#  license information.
#  --------------------------------------------------------------------------
"""Elastic Driver class."""
import json
from datetime import datetime
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ...common.exceptions import MsticpyUserConfigError
from ...common.utility import check_kwargs, export
from ..core.query_defns import Formatters
from .driver_base import DriverBase, DriverProps, QuerySource

__version__ = VERSION
__author__ = "Neil Desai, Ian Hellen"


ELASTIC_CONNECT_ARGS: Dict[str, str] = {
    # TBD - you may not need these - mainly for user
    # help/error messages (see _get_connect_args)
}

_ELASTIC_REQUIRED_ARGS: Dict[str, str] = {
    # TBD
}


@export
class ElasticDriver(DriverBase):
    """Driver to connect and query from Elastic Search."""

    def __init__(self, **kwargs):
        """Instantiate Elastic Driver."""
        super().__init__(**kwargs)
        self.service = None
        self._loaded = True
        self._connected = False
        self._debug = kwargs.get("debug", False)

        self.set_driver_property(
            DriverProps.FORMATTERS,
            {
                Formatters.PARAM_HANDLER: self._custom_param_handler,
                Formatters.DATETIME: self._format_datetime,
                Formatters.LIST: self._format_list,
            },
        )

    def connect(self, connection_str: str = None, **kwargs):
        """
        Connect to Elastic cluster.

        Parameters
        ----------
        connection_str : Optional[str], optional
            Connection string with Splunk connection parameters

        Other Parameters
        ----------------
        kwargs :
            Connection parameters can be supplied as keyword parameters.

        Notes
        -----
        Default configuration is read from the DataProviders/Splunk
        section of msticpyconfig.yaml, if available.

        """
        # cs_dict = self._get_connect_args(connection_str, **kwargs)

        # TBD

        self._connected = True
        print("connected")

    def _get_connect_args(
        self, connection_str: Optional[str], **kwargs
    ) -> Dict[str, Any]:
        """Check and consolidate connection parameters."""
        cs_dict: Dict[str, Any] = {}
        # Fetch any config settings
        cs_dict.update(self._get_config_settings("Elastic"))
        # If a connection string - parse this and add to config
        if connection_str:
            cs_items = connection_str.split(";")
            cs_dict.update(
                {
                    cs_item.split("=")[0].strip(): cs_item.split("=")[1]
                    for cs_item in cs_items
                }
            )
        elif kwargs:
            # if connection args supplied as kwargs
            cs_dict.update(kwargs)
            check_kwargs(cs_dict, list(ELASTIC_CONNECT_ARGS.keys()))

        missing_args = set(_ELASTIC_REQUIRED_ARGS) - cs_dict.keys()
        if missing_args:
            raise MsticpyUserConfigError(
                "One or more connection parameters missing for Elastic connector",
                ", ".join(missing_args),
                f"Required parameters are {', '.join(_ELASTIC_REQUIRED_ARGS)}",
                "All parameters:",
                *[f"{arg}: {desc}" for arg, desc in ELASTIC_CONNECT_ARGS.items()],
                title="no Elastic connection parameters",
            )
        return cs_dict

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query and retrieve results.

        Parameters
        ----------
        query : str
            Elastic query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        kwargs :
            Not used

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        del query_source
        if not self._connected:
            raise self._create_not_connected_err("Elastic")

        # TBD
        # Run query and return results
        return pd.DataFrame()

    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            Query to execute.

        Returns
        -------
        Union[pd.DataFrame,Any]
            A DataFrame (if successful) or
            the underlying provider result if an error occurs.

        """
        raise NotImplementedError(f"Not supported for {self.__class__.__name__}")

    # Parameter Formatting methods
    # If not needed, remove these and remove from self.formatters
    # dict in __init__
    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return f'"{date_time.isoformat(sep=" ")}"'

    @staticmethod
    def _format_list(param_list: Iterable[Any]) -> str:
        """Return formatted list parameter."""
        fmt_list = [f'"{item}"' for item in param_list]
        return ",".join(fmt_list)

    @staticmethod
    def _custom_param_handler(query: str, param_dict: Dict[str, Any]) -> str:
        """Replace parameters in query template for Elastic JSON queries."""
        query_dict = json.loads(query)

        start = param_dict.pop("start", None)
        end = param_dict.pop("end", None)
        if start or end:
            time_range = {
                "range": {"@timestamp": {"format": "strict_date_optional_time"}}
            }
            if start:
                time_range["range"]["@timestamp"]["gte"] = start
            if end:
                time_range["range"]["@timestamp"]["lte"] = end
            query_dict["query"]["bool"]["filter"].append(time_range)

        add_query_items = param_dict.pop("add_query_items", None)
        if add_query_items:
            # "add_query_items" expects additional custom filter parameters
            # as a Python dict (e.g. add_query_items={"match_phrase: {"field": "value"}})
            query_dict["query"]["bool"]["filter"].extend(add_query_items)

        if param_dict:
            filter_terms = [
                {"match_phrase": {field: value}} for field, value in param_dict.items()
            ]
            query_dict["query"]["bool"]["filter"].extend(filter_terms)
        return json.dumps(query_dict, indent=2)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
GreyNoise Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

from typing import Any, ClassVar

from typing_extensions import Self

from ..._version import VERSION
from ..http_provider import APILookupParams
from .ti_http_provider import HttpTIProvider
from .ti_provider_base import ResultSeverity

__version__ = VERSION
__author__ = "Pete Bryan"


class GreyNoise(HttpTIProvider):
    """GreyNoise Lookup."""

    _BASE_URL: ClassVar[str] = "https://api.greynoise.io"

    _QUERIES: ClassVar[dict[str, APILookupParams]] = {
        # Community API
        "ipv4": APILookupParams(
            path="/v3/community/{observable}",
            headers={"key": "{AuthKey}"},
        ),
        # Enterprise API Quick Lookup
        "ipv4-quick": APILookupParams(
            path="/v2/noise/quick/{observable}",
            headers={"key": "{AuthKey}"},
        ),
        # Enterprise API Full Lookup
        "ipv4-full": APILookupParams(
            path="/v2/noise/context/{observable}",
            headers={"key": "{AuthKey}"},
        ),
    }

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return False, ResultSeverity.information, "Not found."
        result = True
        result_dict: dict[str, Any] = {}
        # If community API response extract key elements
        if "riot" in response["RawResult"]:
            result_dict.update(
                {
                    "Classification": response["RawResult"].get("classification"),
                    "Name": response["RawResult"].get("name"),
                    "Last Seen": response["RawResult"].get("last_seen"),
                    "Message": response["RawResult"].get("message"),
                    "Noise": response["RawResult"].get("noise"),
                    "RIOT": response["RawResult"].get("riot"),
                },
            )
        # If enterprise full lookup response extract key elements
        if "actor" in response["RawResult"]:
            result_dict.update(
                {
                    "Classification": response["RawResult"].get("classification"),
                    "First Seen": response["RawResult"].get("first_seen"),
                    "Last Seen": response["RawResult"].get("last_seen"),
                    "Actor": response["RawResult"].get("actor"),
                    "Tags": response["RawResult"].get("tags"),
                    "VPN": response["RawResult"].get("vpn_service", False),
                    "Metadata": response["RawResult"].get("metadata"),
                },
            )
        # If enterprise quick lookup just return raw data is its so small
        if "code" in response["RawResult"]:
            result_dict = response["RawResult"]

        severity = ResultSeverity.information
        if response["RawResult"]["classification"] == "malicious":
            severity = ResultSeverity.high
        return result, severity, result_dict



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Binary Edge Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

from typing import Any, ClassVar

from typing_extensions import Self

from ..._version import VERSION
from ...common.utility import export
from ..http_provider import APILookupParams
from .result_severity import ResultSeverity
from .ti_http_provider import HttpTIProvider

__version__ = VERSION
__author__ = "Pete Bryan"


@export
class BinaryEdge(HttpTIProvider):
    """BinaryEdge Lookup."""

    PROVIDER_NAME: ClassVar[str] = "BinaryEdge"

    _BASE_URL: ClassVar[str] = "https://api.binaryedge.io/"

    _QUERIES: ClassVar[dict[str, APILookupParams]] = {
        "ipv4": APILookupParams(
            path="v2/query/ip/{observable}",
            headers={"X-Key": "{AuthKey}"},
        ),
    }

    _REQUIRED_PARAMS: ClassVar[list[str]] = ["AuthKey"]
    # aliases
    _QUERIES["ipv6"] = _QUERIES["ipv4"]

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """Return the details of the response."""
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return False, ResultSeverity.information, "Not found."

        data: list = response["RawResult"]["events"]
        result_dict: dict[str, Any] = {"IP": response["RawResult"]["query"]}
        open_ports: list[str] = []
        for data_point in data:
            open_ports.append(data_point["port"])
            service_details: dict[str, Any] = {}
            for result in data_point["results"]:
                if "service" in result["result"]["data"]:
                    service_details["Banner"] = result["result"]["data"]["service"][
                        "banner"
                    ]
                if "cert_info" in result["result"]["data"]:
                    service_details["Cert Info"] = result["result"]["data"]["cert_info"]
            result_dict[data_point["port"]] = service_details
        result_dict["Ports"] = open_ports

        return (True, ResultSeverity.information, result_dict)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Check current version against PyPI."""
from importlib.metadata import version

from packaging.version import Version
from packaging.version import parse as parse_version

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def check_version() -> None:
    """Check the current version against latest on PyPI."""
    installed_version: Version = parse_version(__version__)

    # fetch package metadata from PyPI
    distrib_version: str = version("msticpy")

    latest_version: Version = parse_version(distrib_version)

    print(
        "msticpy version",
        "installed:",
        installed_version,
        "latest published:",
        latest_version,
    )
    if installed_version < latest_version:
        print(f"A newer version of msticpy - {latest_version} is available.")
        print("Upgrade with 'pip install --upgrade msticpy'")
    else:
        print("Latest known version is installed.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
HTTP TI Provider base.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

from functools import lru_cache
from json import JSONDecodeError
from typing import TYPE_CHECKING, Any

import pandas as pd
from typing_extensions import Self

from ..._version import VERSION
from ...common.pkg_config import get_http_timeout
from ...common.utility import export
from ..http_provider import HttpProvider
from ..lookup_result import LookupStatus
from .result_severity import ResultSeverity
from .ti_provider_base import TIProvider

if TYPE_CHECKING:
    import httpx
__version__ = VERSION
__author__ = "Ian Hellen"


@export
class HttpTIProvider(TIProvider, HttpProvider):
    """HTTP API Lookup provider base class."""

    def __init__(
        self: HttpTIProvider,
        *,
        timeout: int | None = None,
        ApiID: str | None = None,  # noqa: N803
        AuthKey: str | None = None,  # noqa: N803
        Instance: str | None = None,  # noqa: N803
    ) -> None:
        """Init HttpContextProvider."""
        TIProvider.__init__(self)
        HttpProvider.__init__(
            self,
            timeout=timeout,
            ApiID=ApiID,
            AuthKey=AuthKey,
            Instance=Instance,
        )

    def _run_ti_lookup_query(
        self: Self,
        result: dict[str, Any],
        *,
        query_type: str | None = None,
        timeout: int | None = None,
    ) -> dict[str, Any]:
        """
        Lookup from a value.

        Parameters
        ----------
        result : dict
            Dict result with resolved type and pre-processed item.
            Status is none-zero on failure.
        observable_type : str, optional
            The Type of the observable to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.
        provider_name : str, optional
            Name of the provider to use for the lookup
        timeout : int, optional
            Timeout to use for lookups

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """
        verb, req_params = self._substitute_parms(
            result["SafeIoc"],
            result["IocType"],
            query_type,
        )
        if verb == "GET":
            response: httpx.Response = self._httpx_client.get(
                **req_params,
                timeout=get_http_timeout(timeout=timeout),
            )
        elif verb == "POST":
            response = self._httpx_client.post(
                **req_params,
                timeout=get_http_timeout(timeout=timeout),
            )
        else:
            err_msg: str = f"Unsupported verb {verb}"
            raise NotImplementedError(err_msg)

        result["Status"] = response.status_code
        result["Reference"] = req_params["url"]
        if response.is_success:
            try:
                result["RawResult"] = response.json()
                result["Result"], severity, result["Details"] = self.parse_results(
                    result,
                )
            except JSONDecodeError:
                result[
                    "RawResult"
                ] = f"""There was a problem parsing results from this lookup:
                                    {response.text}"""
                result["Result"] = False
                severity = ResultSeverity.information
                result["Details"] = {}
            if isinstance(severity, ResultSeverity):
                result["Severity"] = severity.name
            result["Severity"] = ResultSeverity.parse(severity).name
            result["Status"] = LookupStatus.OK.value
        else:
            result["RawResult"] = str(response)
            result["Result"] = False
            result["Details"] = self._response_message(result["Status"])
        return result

    @lru_cache(maxsize=256)
    def lookup_ioc(  # noqa: PLR0913
        self: Self,
        ioc: str,
        ioc_type: str | None = None,
        query_type: str | None = None,
        *,
        provider_name: str | None = None,
        timeout: int = 120,
    ) -> pd.DataFrame:
        """
        Lookup from a value.

        Parameters
        ----------
        ioc : str
            ioc to lookup
        ioc_type : str, optional
            The Type of the ioc to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the ioc
            will be returned.
        provider_name : str, optional
            Name of the provider to query for the lookup
        timeout : str, optional
            Timeout for lookup queries

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """
        result: dict[str, Any] = self._check_ioc_type(
            ioc,
            ioc_type,
            query_subtype=query_type,
        )
        result["Provider"] = provider_name or self.__class__.__name__

        req_params: dict[str, Any] = {}
        try:
            self._run_ti_lookup_query(
                result=result,
                query_type=query_type,
                timeout=timeout,
            )
        except (
            LookupError,
            JSONDecodeError,
            NotImplementedError,
            ConnectionError,
        ) as err:
            self._err_to_results(result, err)
            if not isinstance(err, LookupError):
                url: str | None = req_params.get("url") if req_params else None
                result["Reference"] = url
        return pd.DataFrame([result])



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebook utility functions."""
# pickle only used here for storing data.
import pickle  # nosec
from base64 import b64encode
from typing import Any, Iterable, Optional, Union

from IPython import get_ipython
from IPython.display import HTML, DisplayHandle, display

from ..._version import VERSION
from .types import export

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name
@export
def md(
    string: str,
    styles: Union[str, Iterable[str]] = None,
    disp_id: Optional[Union[bool, DisplayHandle]] = None,
) -> Optional[DisplayHandle]:
    """
    Display a string as Markdown with optional style.

    Parameters
    ----------
    string : str
        The string to display
    styles : Union[str, Iterable[str]], optional
        A style mnemonic or collection of styles. If multiple styles,
        these can be supplied as an interable of strings or a comma-separated
        string, by default None
    disp_id : Optional[Union[bool, DisplayHandle]], optional
        If True, the function will return a display handle that can be re-used
        in subsequent calls to update the display object.
        If this is previously-created display handle, this is used as the
        target display object to update it with the content of this call,
        by default None

    Returns
    -------
    DisplayHandle
        A handle to the display object that can be used to update the
        contents.

    """
    style_str = ""
    if isinstance(styles, str):
        if "," in styles:
            styles = [style.strip() for style in styles.split(",")]
        else:
            style_str = _F_STYLES.get(styles, "")
    if isinstance(styles, list):
        style_str = ";".join(_F_STYLES.get(style, "") for style in styles)
    content = HTML(f"<p style='{style_str}'>{string}</p>")

    if isinstance(disp_id, bool) and disp_id:
        return display(content, display_id=True)
    if isinstance(disp_id, DisplayHandle):
        return disp_id.update(content)
    display(content)
    return None


# pylint: enable=invalid-name


@export
def md_warn(string: str, disp_id: Optional[DisplayHandle] = None):
    """
    Return string as a warning - orange text prefixed by "Warning".

    Parameters
    ----------
    string : str
        The warning message.
    disp_id : Optional[DisplayHandle], optional
        If True, the function will return a display handle that can be re-used
        in subsequent calls to update the display object.
        If this is previously-created display handle, this is used as the
        target display object to update it with the content of this call,
        by default None

    Returns
    -------
    DisplayHandle
        A handle to the display object that can be used to update the
        contents.

    """
    return md(f"Warning: {string}", "bold, orange, large", disp_id)


@export
def md_error(string: str, disp_id: Optional[DisplayHandle] = None):
    """
    Return string as an error - red text prefixed by "Error".

    Parameters
    ----------
    string : str
        The error message.
    disp_id : Optional[Union[bool, DisplayHandle]], optional
        If True, the function will return a display handle that can be re-used
        in subsequent calls to update the display object.
        If this is previously-created display handle, this is used as the
        target display object to update it with the content of this call,
        by default None

    """
    return md(f"Error: {string}", "bold, orange, large", disp_id)


# Styles available to use in the above Markdown tools.
_F_STYLES = {
    "bold": "font-weight: bold",
    "italic": "font-style: italic",
    "red": "color: red",
    "green": "color: green",
    "blue": "color: blue",
    "large": "font-size: 130%",
    "heading": "font-size: 200%",
    "outline": "border: solid; padding: 5pt",
}


@export
def is_ipython(notebook: bool = False) -> bool:
    """
    Return True if running in IPython environment.

    Parameters
    ----------
    notebook : bool, optional
        If notebook is true this will only return true if running
        in a Jupyter notebook.

    Returns
    -------
    bool
        True if running in IPython environment,
        otherwise False

    """
    return (
        get_ipython() and type(get_ipython()).__name__.startswith("ZMQ")
        if notebook
        else bool(get_ipython())
    )


_CODE_CELL_TEMPLATE = """#########################################
# Run this cell to restore cached data to
# the object "{var_name}"
#########################################

from base64 import b64decode
import pickle

## Store dynamic summaries as base64 byte string
summary_data = {encoded_bytes}

# decode and unpickle the summaries
{var_name} = pickle.loads(b64decode(summary_data))
{var_name}
"""


@export
def save_obj_to_cell(obj: Any, var_name: str):
    """
    Save a pickle-able object to a new cell.

    Parameters
    ----------
    obj : Any
        The object to be stored. Must be pickle-able.
    var_name : str
        The variable name use to restore the object
        in the new cell.

    Raises
    ------
    TypeError
        If the object does not support pickling.

    Notes
    -----
    Saves `obj` as picked, base64-encoded blob to
    a new notebook cell.

    """
    try:
        picked_obj = pickle.dumps(obj)
    except pickle.PickleError as err:
        raise TypeError(f"Object of type {type(obj)} cannot be picked.") from err
    encoded_bytes = b64encode(picked_obj)
    cell_text = _CODE_CELL_TEMPLATE.format(
        encoded_bytes=encoded_bytes,
        var_name=var_name,
    )
    shell = get_ipython()
    # create a new cell using `cell_code` as the code contents.
    shell.set_next_input(cell_text)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Azure Cloud Mappings."""
import contextlib
from functools import lru_cache
from typing import Any, Dict, List, Optional

import httpx

from .._version import VERSION
from ..common import pkg_config as config
from ..common.exceptions import MsticpyAzureConfigError
from ..common.pkg_config import get_http_timeout
from ..common.utility.package import unit_testing
from .cloud_mappings_offline import cloud_mappings_offline

__version__ = VERSION
__author__ = "Pete Bryan"

CLOUD_MAPPING = {
    "global": "https://management.azure.com/",
    "usgov": "https://management.usgovcloudapi.net/",
    "cn": "https://management.chinacloudapi.cn/",
}

CLOUD_ALIASES = {
    "public": "global",
    "global": "global",
    "gov": "usgov",
    "usgov": "usgov",
    "china": "cn",
    "cn": "cn",
}

_DEFENDER_MAPPINGS = {
    "global": "https://api.securitycenter.microsoft.com/",
    "us": "https://api-us.securitycenter.microsoft.com/",
    "eu": "https://api-eu.securitycenter.microsoft.com/",
    "uk": "https://api-uk.securitycenter.microsoft.com/",
    "gcc": "https://api-gcc.securitycenter.microsoft.us/",
    "gcc-high": "https://api-gov.securitycenter.microsoft.us/",
    "dod": "https://api-gov.securitycenter.microsoft.us/",
}

_M365D_MAPPINGS = {
    "global": "https://api.security.microsoft.com/",
    "us": "https://api-us.security.microsoft.com/",
    "eu": "https://api-eu.security.microsoft.com/",
    "uk": "https://api-uk.security.microsoft.com/",
    "gcc": "https://api-gcc.security.microsoft.us",
    "gcc-high": "https://api-gov.security.microsoft.us",
    "dod": "https://api-gov.security.microsoft.us",
}


def format_endpoint(endpoint: str) -> str:
    """Format an endpoint with "/" if needed ."""
    return endpoint if endpoint.endswith("/") else f"{endpoint}/"


@lru_cache(maxsize=None)
def get_cloud_endpoints(
    cloud: str, resource_manager_url: Optional[str] = None
) -> Dict[str, Any]:
    """
    Get the cloud endpoints for a specific cloud.

    If resource_manager_url is supplied, it will be used instead of the cloud name.

    Parameters
    ----------
    cloud : str
        The name of the cloud to get endpoints for.
    resource_manager_url : str, optional
        The resource manager url for a cloud.
        Can be used to get all endpoints for a specific cloud. Defaults to None.

    Returns
    -------
    dict
        A dictionary of endpoints and suffixes for the specified cloud/resource_manager_url.

    Raises
    ------
    MsticpyAzureConfigError
        If the cloud name is not valid.

    """
    response = None
    if resource_manager_url:
        response = get_cloud_endpoints_by_resource_manager_url(resource_manager_url)
    else:
        response = get_cloud_endpoints_by_cloud(cloud)
    if response:
        return response

    raise MsticpyAzureConfigError(
        f"Error retrieving endpoints for Cloud: {cloud}",
        f"Resource Manager Url: {resource_manager_url}.",
    )


def get_cloud_endpoints_by_cloud(cloud: str) -> Dict[str, Any]:
    """
    Get the cloud endpoints for a specific cloud.

    Parameters
    ----------
    cloud : str
        The name of the cloud to get endpoints for.

    Returns
    -------
    Dict
        Contains endpoints and suffixes for a specific cloud.

    """
    resource_manager_url = CLOUD_MAPPING.get(CLOUD_ALIASES.get(cloud, "global"))
    return get_cloud_endpoints_by_resource_manager_url(resource_manager_url)  # type: ignore


def get_cloud_endpoints_by_resource_manager_url(
    resource_manager_url: str,
) -> Dict[str, Any]:
    """
    Get the cloud endpoints for a specific resource manager url.

    Parameters
    ----------
    resource_manager_url : str
        The resource manager url to get endpoints for.

    Returns
    -------
    Dict
        Contains endpoints and suffixes for a specific cloud.

    """
    if unit_testing():
        return cloud_mappings_offline["global"]
    f_resource_manager_url = format_endpoint(resource_manager_url)
    endpoint_url = f"{f_resource_manager_url}metadata/endpoints?api-version=latest"
    try:
        resp = httpx.get(endpoint_url, timeout=get_http_timeout())
        if resp.status_code == 200:
            return resp.json()

    except httpx.RequestError:
        cloud = next(
            (
                key
                for key, val in CLOUD_MAPPING.items()
                if val == f_resource_manager_url
            ),
            "global",
        )
        return cloud_mappings_offline[cloud]

    return cloud_mappings_offline["global"]


def get_azure_config_value(key, default):
    """Get a config value from Azure section."""
    with contextlib.suppress(KeyError):
        az_settings = config.get_config("Azure", {})
        if az_settings and key in az_settings:
            return az_settings[key]
    return default


def default_auth_methods() -> List[str]:
    """Get the default (all) authentication options."""
    return get_azure_config_value(
        "auth_methods", ["env", "msi", "vscode", "cli", "powershell", "devicecode"]
    )


def get_defender_endpoint(cloud: str) -> str:
    """Get the URI of the applicable Defender for Endpoint API."""
    return _DEFENDER_MAPPINGS[cloud.casefold()]


def get_m365d_endpoint(cloud: str) -> str:
    """Get the URI of the applicable Defender for Endpoint API."""
    return _M365D_MAPPINGS[cloud]


def get_m365d_login_endpoint(cloud: str) -> str:
    """Get M365 login URL."""
    if cloud in {"gcc-high", "dod"}:
        return "https://login.microsoftonline.us/"
    return AzureCloudConfig().authority_uri


class AzureCloudConfig:
    """Azure Cloud configuration."""

    def __init__(
        self,
        cloud: Optional[str] = None,
        tenant_id: Optional[str] = None,
        resource_manager_url: Optional[str] = None,
    ):
        """
        Initialize AzureCloudConfig from `cloud` or configuration.

        Parameters
        ----------
        cloud : str, optional
            The cloud to retrieve configuration for. If not supplied,
            the cloud ID is read from configuration. If this is not available,
            it defaults to 'global'.
        tenant_id : str, optional
            The tenant to authenticate against. If not supplied,
            the tenant ID is read from configuration, or the default tenant
            for the identity.
        resource_manager_url : str, optional
            The resource manager URL to use. If not supplied,
            the URL is based on the cloud name within the configuration
            or defaults to 'https://management.azure.com/'.

        """
        self.cloud = cloud or get_azure_config_value("cloud", "global")
        self.tenant_id = tenant_id or get_azure_config_value("tenant_id", None)
        self.auth_methods = default_auth_methods()
        self.resource_manager_url = resource_manager_url or get_azure_config_value(
            "resource_manager_url", None
        )
        self.endpoints = get_cloud_endpoints(self.cloud, self.resource_manager_url)

    @property
    def cloud_names(self) -> List[str]:
        """Return a list of current cloud names."""
        return list(CLOUD_MAPPING.keys())

    @staticmethod
    def resolve_cloud_alias(
        alias,
    ) -> Optional[str]:
        """Return match of cloud alias or name."""
        alias_cf = alias.casefold()
        aliases = {alias.casefold(): cloud for alias, cloud in CLOUD_ALIASES.items()}
        if alias_cf in aliases:
            return aliases[alias_cf]
        return alias_cf if alias_cf in aliases.values() else None

    @property
    def suffixes(self) -> Dict[str, str]:
        """
        Get CloudSuffixes class an Azure cloud.

        Returns
        -------
        dict
            Dict of cloud endpoint suffixes.

        Raises
        ------
        MsticpyAzureConfigError
            If the cloud name is not valid.

        """
        return self.endpoints.get("suffixes", {})

    @property
    def token_uri(self) -> str:
        """Return the resource manager token URI."""
        rm_url = self.resource_manager_url or self.resource_manager
        rm_url = format_endpoint(rm_url)
        return f"{rm_url}.default"

    @property
    def authority_uri(self) -> str:
        """Return the AAD authority URI."""
        return format_endpoint(
            self.endpoints.get("authentication", {}).get("loginEndpoint")
        )

    @property
    def log_analytics_uri(self) -> str:
        """Return the AAD authority URI."""
        return format_endpoint(
            self.endpoints.get("logAnalyticsResourceId")  # type: ignore
        )

    @property
    def resource_manager(self) -> str:
        """Return the resource manager URI."""
        return format_endpoint(self.endpoints.get("resourceManager"))  # type: ignore
